from io import StringIO
import json
import pytest

import datalad
from datalad.api import (
    credentials,
    download,
)
from datalad_next.tests import (
    assert_result_count,
    assert_status,
)
from datalad_next.utils import chpwd

from datalad_next.utils import CredentialManager


@pytest.fixture
def hbsurl(httpbin):
    # shortcut for the standard URL
    return httpbin["standard"]


test_cred = ('dltest-my&=http', 'datalad', 'secure')


@pytest.fixture
def hbscred(hbsurl):
    return (
        'hbscred',
        dict(user='mike', secret='dummy', type='user_password',
             realm=f'{hbsurl}/Fake Realm'),
    )


def test_download(tmp_path, http_server, no_result_rendering):
    wdir = tmp_path
    srvurl = http_server.url
    (http_server.path / 'testfile.txt').write_text('test')

    # simple download, taking the target filename from the URL
    # single-pass hashing with two algorithms
    with chpwd(wdir):
        res = download(f'{srvurl}testfile.txt',
                       hash=['md5', 'SHA256'],
                       return_type='item-or-list')

    assert (wdir / 'testfile.txt').read_text() == 'test'
    # keys for hashes keep user-provided capitalization
    assert res['md5'] == '098f6bcd4621d373cade4e832627b4f6'
    assert res['SHA256'] == \
        '9f86d081884c7d659a2feaa0c55ad015a3bf4f1b2b0b822cd15d6c15b0f00a08'

    # doing it again fails due to overwrite detection
    with chpwd(wdir):
        assert_result_count(
            download(f'{srvurl}testfile.txt', on_failure='ignore'),
            1, status='error', error_message='target path already exists')

    # works with explicit alternative filename
    with chpwd(wdir):
        download(f'{srvurl}testfile.txt testfile2.txt')

    assert (wdir / 'testfile2.txt').read_text() == 'test'

    # non-existing download source
    assert_result_count(
        download(f'{srvurl}nothere', on_failure='ignore'),
        1, status='error', message='download failure')


def test_download_invalid_calls(monkeypatch, no_result_rendering):
    # unsupported url scheme, only detected when actually calling
    # a handler inside, hence error result
    assert_result_count(
        download('dummy://mike/file', on_failure='ignore'),
        1,
        status='error',
        message='unsupported URL (custom URL handlers can be declared '
        'via DataLad configuration)',
    )
    # no target path derivable
    # immediate error, when all information is readily available
    with pytest.raises(ValueError):
        download('http://example.com')
    # deferred error result when a generator is gathering batch-mode
    # items at runtime
    monkeypatch.setattr('sys.stdin', StringIO('http://example.com'))
    assert_result_count(
        download(
            '-',
            on_failure='ignore',
        ),
        1, status='impossible')


def test_download_auth(
        tmp_path, credman, http_credential, http_server_with_basicauth,
        no_result_rendering):
    credman.set(**http_credential)
    wdir = tmp_path
    srvurl = http_server_with_basicauth.url
    srvpath = http_server_with_basicauth.path
    (srvpath / 'testfile.txt').write_text('test')

    # we have a credential, but there is nothing to discover from
    # that we should use it for this request
    assert_result_count(
        download(f'{srvurl}nothere', on_failure='ignore'),
        1, status='error', message='download failure')

    # amend the test credential with the realm of the test server
    credman = CredentialManager(datalad.cfg)
    credman.set(test_cred[0], realm=f'{srvurl}Protected')

    # now it must be able to auto-detect it
    download({f'{srvurl}testfile.txt': wdir / 'download1.txt'})
    assert (wdir / 'download1.txt').read_text() == 'test'


auth_ok_response = {"authenticated": True, "user": "mike"}


def test_download_basic_auth(credman, capsys, hbscred, hbsurl,
                             no_result_rendering):
    credman.set(hbscred[0], **hbscred[1])
    # consume stdout to make test self-contained
    capsys.readouterr()
    download(
        {f'{hbsurl}/basic-auth/mike/dummy': '-'})
    assert json.loads(capsys.readouterr().out) == auth_ok_response


def test_download_bearer_token_auth(credman, capsys, hbsurl):
    credman.set('dummy', realm=f'{hbsurl}/', type='token', secret='very')
    # consume stdout to make test self-contained
    capsys.readouterr()
    download(
        {f'{hbsurl}/bearer': '-'})
    assert json.loads(capsys.readouterr().out) == {
        'authenticated': True,
        'token': 'very',
    }


def test_download_digest_auth(credman, capsys, hbscred, hbsurl,
                              no_result_rendering):
    credman.set(hbscred[0],
                **dict(hbscred[1],
                       realm=f'{hbsurl}/me@kennethreitz.com'))
    # consume stdout to make test self-contained
    capsys.readouterr()
    for url_suffix in (
            '/digest-auth/auth/mike/dummy',
            # non-default algorithm
            '/digest-auth/auth/mike/dummy/SHA-256',
    ):
        download({f'{hbsurl}{url_suffix}': '-'})
        assert json.loads(capsys.readouterr().out) == auth_ok_response
        # repeated reads do not accumulate
        assert capsys.readouterr().out == ''


def test_download_explicit_credential(credman, capsys, hbscred, hbsurl):
    # the provided credential has the wrong 'realm' for auto-detection.
    # but choosing it explicitly must put things to work
    credman.set(hbscred[0], **hbscred[1])
    # consume stdout to make test self-contained
    capsys.readouterr()
    download({f'{hbsurl}/digest-auth/auth/mike/dummy': '-'},
             credential=hbscred[0])
    assert json.loads(capsys.readouterr().out) == auth_ok_response


def test_download_auth_after_redirect(credman, capsys, hbscred, hbsurl):
    credman.set(hbscred[0], **hbscred[1])
    # consume stdout to make test self-contained
    capsys.readouterr()
    download(
        {f'{hbsurl}/redirect-to?url={hbsurl}/basic-auth/mike/dummy': '-'})
    assert json.loads(capsys.readouterr().out) == auth_ok_response


def test_download_no_credential_leak_to_http(credman, capsys, hbscred, httpbin,
                                             no_result_rendering):
    credman.set(hbscred[0], **hbscred[1])
    redirect_url = f'{httpbin["http"]}/basic-auth/mike/dummy'
    res = download(
        # redirect from https to http, must drop provided credential
        # to avoid leakage
        {f'{httpbin["https"]}/redirect-to?url={redirect_url}': '-'},
        credential=hbscred[0],
        on_failure='ignore')
    assert_status('error', res)
    assert '401' in res[0]['error_message']
    assert f' {redirect_url}' in res[0]['error_message']
    # do the same again, but without the explicit credential,
    # also must not work
    # this is not the right test, though. What would be suitable
    # is an authenticated request that then redirects
    res = download(
        # redirect from https to http, must drop provided credential
        # to avoid leakage
        {f'{httpbin["https"]}/redirect-to?url={redirect_url}': '-'},
        on_failure='ignore')
    assert_status('error', res)


def test_download_new_bearer_token(
        tmp_keyring, capsys, hbsurl, datalad_interactive_ui,
        no_result_rendering):
    ui = datalad_interactive_ui
    ui.staged_responses.extend([
        'token123',
        # after download, it asks for a name
        'dataladtest_test_download_new_bearer_token',
    ])
    try:
        download({f'{hbsurl}/bearer': '-'})
        # and it was saved under this name
        assert_result_count(
            credentials(
                'get',
                name='dataladtest_test_download_new_bearer_token'),
            1, cred_secret='token123', cred_type='token',
        )
    finally:
        credentials(
            'remove',
            name='dataladtest_test_download_new_bearer_token',
        )


def test_download_new_bearer_token_nosave(
        capsys, hbsurl, datalad_interactive_ui,
        no_result_rendering):
    ui = datalad_interactive_ui
    ui.staged_responses.extend([
        'datalad_uniquetoken123',
        # after download, it asks for a name, but skip to save
        'skip',
    ])
    download({f'{hbsurl}/bearer': '-'})
    # and it was saved under this name
    assert_result_count(
        credentials('query', dict(secret='datalad_uniquetoken123')),
        0,
    )


# make sure a 404 is easily discoverable
# https://github.com/datalad/datalad/issues/6545
def test_download_404(hbsurl, no_result_rendering):
    assert_result_count(
        download(f'{hbsurl}/status/404', on_failure='ignore'),
        1, status_code=404, status='error')


def test_downloadurl(tmp_path, no_result_rendering):
    (tmp_path / 'src').mkdir()
    dst_path = tmp_path / 'dst'
    dst_path.mkdir()
    testfile = tmp_path / 'src' / 'myfile.txt'
    testfile.write_text('some content')

    res = download(
        {testfile.as_uri(): dst_path / 'target.txt'},
        hash=['md5'],
        return_type='item-or-list')
    assert_result_count(res, 1, md5='9893532233caff98cd083a116b013c0b')



from __future__ import annotations

import logging
from abc import (
    ABCMeta,
    abstractmethod,
)
from collections import deque
from collections.abc import Generator
from random import randint

from datasalad.iterable_subprocess.iterable_subprocess import OutputFrom
from datasalad.itertools import align_pattern


__all__ = [
    'FixedLengthResponseGenerator',
    'FixedLengthResponseGeneratorPosix',
    'FixedLengthResponseGeneratorPowerShell',
    'ShellCommandResponseGenerator',
    'VariableLengthResponseGenerator',
    'VariableLengthResponseGeneratorPosix',
    'VariableLengthResponseGeneratorPowerShell',
]


lgr = logging.getLogger('datalad.ext.next.shell.protocol')


class ShellCommandResponseGenerator(Generator, metaclass=ABCMeta):
    """An abstract class the specifies the minimal functionality of a response generator

    Subclasses of this class can be used to implement operation-specific,
    shell-specific or OS-specific details of the command execution and the
    command output parsing.

    The return code is available in the ``returncode``-attribute, the
    stderr-output is available in the ``stderr_deque``-attribute (a
    ``deque``-instance), of instances of this class.
    """
    def __init__(self, stdout_gen: Generator, stderr_deque: deque) -> None:
        self.stdout_gen = stdout_gen
        self.stderr_deque = stderr_deque
        self.state: str | int = 'output'
        self.returncode_chunk = b''
        self.returncode: int | None = None

    @staticmethod
    def _get_number_and_newline(chunk, iterable) -> tuple[int, bytes]:
        """Help that reads a trailing number and a newline from a chunk

        Parameters
        ----------
        chunk : bytes
            An chunk of bytes that should contain the number and the newline.
        iterable : Iterable
            An iterable that will be used to extend ``chunk`` if no
            newline is found in ``chunk``.

        Returns
        -------
        int
            A tuple that contains the number that was found in the chunk and
            the trailing portion of the chunk that was not parsed.
        """
        while b'\n' not in chunk:
            lgr.log(5, 'completing number chunk')
            chunk += next(iterable)
        digits, trailing = chunk.split(b'\n', 1)
        return int(digits), trailing

    @abstractmethod
    def send(self, _) -> bytes:
        """Deliver the next part of generated output

        Whenever the response generator is iterated over, this method is called
        and should deliver the next part of the command output or raise
        ``StopIteration`` if the command has finished.
        """
        raise NotImplementedError

    @abstractmethod
    def get_final_command(self, command: bytes) -> bytes:
        """Return a final command list that executes ``command``

        This method should return a "final" command-pipeline that executes
        ``command`` and generates the output structure that the response
        generator expects. This structure will typically be parsed in the
        implementation of :meth:`send`.

        This method is usually only called by
        :meth:`ShellCommandExecutor.__call__`.
        """
        raise NotImplementedError

    def throw(self, typ, val=..., tb=...):  # pragma: no cover
        return super().throw(typ, val, tb)


class VariableLengthResponseGenerator(ShellCommandResponseGenerator, metaclass=ABCMeta):
    """Response generator that handles outputs of unknown length

    This response generator is used to execute a command that will result in an
    output of unknown length, e.g. ``ls``. The final command list it creates
    will execute the command and print a random end-marker and the return code
    after the output of the command. The :meth:`send`-method of this class uses
    the end-marker to determine then end of the command output.
    """
    def __init__(self,
                 stdout: OutputFrom,
                 ) -> None:
        self.end_marker = _create_end_marker()
        self.stream_marker = self.end_marker + b'\n'
        self.plain_stdout = stdout
        super().__init__(
            align_pattern(stdout, self.stream_marker),
            stdout.stderr_deque
        )

    def send(self, _) -> bytes:
        if self.state == 'output':
            chunk = next(self.stdout_gen)
            if self.stream_marker in chunk:
                self.state = 'returncode'
                chunk, self.returncode_chunk = chunk.split(self.stream_marker)
                if chunk:
                    return chunk
            else:
                return chunk

        if self.state == 'returncode':
            self.returncode, trailing = self._get_number_and_newline(
                self.returncode_chunk,
                self.plain_stdout,
            )
            if trailing:
                lgr.warning(
                    'unexpected output after return code: %s',
                    repr(trailing))
            self.state = 'exhausted'

        if self.state == 'exhausted':
            self.state = 'output'
            raise StopIteration()

        raise RuntimeError(f'unknown state: {self.state}')

    @property
    @abstractmethod
    def zero_command(self) -> bytes:
        """Return a command that functions as "zero command" """
        raise NotImplementedError


class VariableLengthResponseGeneratorPosix(VariableLengthResponseGenerator):
    """A variable length response generator for POSIX shells"""
    def __init__(self, stdout):
        """
        Parameters
        ----------
        stdout : OutputFrom
            A generator that yields output from a shell. Usually the object
            that is returned by :func:`iter_proc`.
        """
        super().__init__(stdout)

    def get_final_command(self, command: bytes) -> bytes:
        """Return a command list that executes ``command`` and prints the end-marker

        The POSIX version for variable length response generators.

        This method is usually only called by
        :meth:`ShellCommandExecutor.__call__`.
        """
        return (
            command + b' ; x=$?; echo -e -n "' + self.end_marker + b'\\n"; echo $x\n'
        )

    @property
    def zero_command(self) -> bytes:
        return b'test 0 -eq 0'


class VariableLengthResponseGeneratorPowerShell(VariableLengthResponseGenerator):
    """A variable length response generator for PowerShell shells"""
    def __init__(self, stdout):
        """
        Parameters
        ----------
        stdout : OutputFrom
            A generator that yields output from a shell. Usually the object
            that is returned by :func:`iter_proc`.
        """
        super().__init__(stdout)

    def get_final_command(self, command: bytes) -> bytes:
        """Return a command list that executes ``command`` and prints the end-marker

        The PowerShell version for variable length response generators.

        This method is usually only called by
        :meth:`ShellCommandExecutor.__call__`.
        """
        # TODO: check whether `command` sets `$LASTEXITCODE` and assign that
        #  to `$x`, iff set.
        return (
            b'$x=0; try {' + command + b'} catch { $x=1 }\n'
            + b'Write-Host -NoNewline ' + self.end_marker + b'`n$x`n\n'
        )

    @property
    def zero_command(self) -> bytes:
        return b'Write-Host hello'


class FixedLengthResponseGenerator(ShellCommandResponseGenerator, metaclass=ABCMeta):
    """Response generator for efficient handling of outputs of known length

    This response generator is used to execute commands that have an output of
    known length. The final command list it creates will execute the command and
    print the return code followed by a newline.

    The :meth:`send`-method of this response generator will read the specified
    number of bytes and a trailing return code. This is more performant than
    scanning the output for an end-marker.
    """
    def __init__(self,
                 stdout: OutputFrom,
                 length: int,
                 ) -> None:
        """
        Parameters
        ----------
        stdout : OutputFrom
            A generator that yields output from a shell. Usually the object
            that is returned by :func:`iter_proc`.
        length : int
            The length (in bytes) of the output that a command will generate.
        """
        super().__init__(stdout, stdout.stderr_deque)
        self.length = length
        self.read = 0

    def send(self, _) -> bytes:
        if self.state == 'output':
            chunk = next(self.stdout_gen)
            self.read += len(chunk)
            if self.read >= self.length:
                self.state = 'returncode'
                excess = self.read - self.length
                if excess > 0:
                    chunk, self.returncode_chunk = chunk[:-excess], chunk[-excess:]
                else:
                    self.returncode_chunk = b''
                if chunk:
                    return chunk
            else:
                return chunk

        if self.state == 'returncode':
            self.returncode, trailing = self._get_number_and_newline(
                self.returncode_chunk,
                self.stdout_gen,
            )
            if trailing:
                lgr.warning(
                    'unexpected output after return code: %s',
                    repr(trailing))
            self.state = 'exhausted'

        if self.state == 'exhausted':
            self.state = 'output'
            raise StopIteration()

        raise RuntimeError(f'unknown state: {self.state}')


class FixedLengthResponseGeneratorPosix(FixedLengthResponseGenerator):
    def get_final_command(self, command: bytes) -> bytes:
        """Return a final command list for a command with a fixed length output

        The POSIX version for fixed length response generators.

        This method is usually only called by
        :meth:`ShellCommandExecutor.__call__`.
        """
        return command + b' ; echo $?\n'


class FixedLengthResponseGeneratorPowerShell(FixedLengthResponseGenerator):
    def get_final_command(self, command: bytes) -> bytes:
        """Return a final command list for a command with a fixed length output

        The PowerShell version for fixed length response generators.

        This method is usually only called by
        :meth:`ShellCommandExecutor.__call__`.
        """
        return (
            b'$x=0; try {' + command + b'} catch { $x=1 }\n'
            + b'Write-Host -NoNewline $x`n\n'
        )


def _create_end_marker() -> bytes:
    """ Create a hopefully unique marker for the shell """
    # The following line is marked with `nosec` because `randint` is only
    # used to diversify markers, not for cryptographic purposes.
    marker_id = f'{randint(1000000000, 9999999999)}'.encode()  # nosec
    fixed_part = b'----datalad-end-marker-'
    return fixed_part + marker_id + fixed_part[::-1]



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
import logging
from pathlib import (
    Path,
    PurePath,
)

from datalad_next.datasets import Dataset
from datalad_next.exceptions import CapturedException


# TODO Could be `StrEnum`, came with PY3.11
class CommandResultStatus(Enum):
    """Enumeration of possible statuses of command results
    """
    ok = 'ok'
    notneeded = 'notneeded'
    impossible = 'impossible'
    error = 'error'


# which status is a success , which is failure
success_status_map = {
    'ok': 'success',
    'notneeded': 'success',
    'impossible': 'failure',
    'error': 'failure',
}


# We really want this to be `kw_only=True`, but cannot, because it only
# came with PY3.10
# Until this can be enabled, we cannot have additional _required_ properties
# coming from derived classes. Instead, we have to make any and all
# additional properties optional (with default None), because also in this
# base class we do define optional ones (and it makes no sense not to do
# that either).
#@dataclass(kw_only=True)
@dataclass
class CommandResult:
    """Base data class for result records emitted by DataLad commands.

    Historically, such results records have taken the form of a Python
    ``dict``. This class provides some API for its instances to be
    compatible with legacy code that expects a ``dict``.

    .. seealso::

      https://docs.datalad.org/design/result_records.html
    """
    # TODO implement post_init and possibly check for validated of
    # some arguments (e.g. status is a valid value). Maybe do all of that
    # conditional on some config flag that could be set during test
    # execution

    # mandatory as per
    # http://docs.datalad.org/design/result_records.html#mandatory-fields
    action: str
    """A string label identifying which type of operation a result is
    associated with. Labels must not contain white space. They should be
    compact, and lower-cases, and use ``_`` (underscore) to separate words in
    compound labels.
    """
    status: CommandResultStatus
    """This field indicates the nature of a result in terms of four
    categories, identified by a :class:`CommandResultStatus` value.
    The result status is used by user communication, but also for decision
    making on the overall success or failure of a command operation.
    """
    path: str | Path
    """An *absolute* path describing the local entity a result is associated
    with (the subject of the result record). Paths must be platform-specific
    (e.g., Windows paths on Windows, and POSIX paths on other operating
    systems). When a result is about an entity that has no meaningful relation
    to the local file system (e.g., a URL to be downloaded), the ``path`` value
    should be determined with respect to the potential impact of the result
    on any local entity (e.g., a URL downloaded to a local file path, a local
    dataset modified based on remote information).
    """
    # optional
    # TODO complete documentation of all members
    message: str | tuple | None = None
    exception: CapturedException | None = None
    error_message: str | tuple | None = None
    type: str | None = None
    logger: logging.Logger | None = None
    refds: str | Path | Dataset = None

    # any and all of the code below makes it possible to feed such result
    # instances through the datalad-core result processing loop (which
    # expects results to be dicts with string keys and (most) values to
    # be string only also.
    def __contains__(self, key: str) -> bool:
        return hasattr(self, key)

    def __getitem__(self, key: str):
        return self._stringify4legacy(getattr(self, key))

    def get(self, key, default=None):
        return self._stringify4legacy(getattr(self, key, default))

    def pop(self, key, default=None):
        item = getattr(self, key, default)
        if hasattr(self, key):
            setattr(self, key, None)
        return self._stringify4legacy(item)

    def items(self):
        for k, v in self.__dict__.items():
            yield k, self._stringify4legacy(v)

    def _stringify4legacy(self, val):
        if isinstance(val, PurePath):
            return str(val)
        elif isinstance(val, Dataset):
            return val.path
        elif issubclass(getattr(val, '__class__', None), Enum):
            return val.value
        return val



from pathlib import Path

from datalad.dataset.gitrepo import GitRepo as LeanGitRepo
from datalad.support.annexrepo import AnnexRepo as LegacyAnnexRepo


class LeanAnnexRepo(LegacyAnnexRepo):
    """git-annex repository representation with a minimized API

    This is a companion of :class:`LeanGitRepo`. In the same spirit, it
    restricts its API to a limited set of method that extend
    :class:`LeanGitRepo`.

    """
    #CA .. autosummary::

    #CA    call_annex
    #CA    call_annex_oneline
    #CA    call_annex_success
    # list of attributes permitted in the "lean" API. This list extends
    # the API of LeanGitRepo
    # TODO extend whitelist of attributes as necessary
    _lean_attrs = [
        #CA # these are the ones we intend to provide
        #CA 'call_annex',
        #CA 'call_annex_oneline',
        #CA 'call_annex_success',
        # and here are the ones that we need to permit in order to get them
        # to run
        '_check_git_version',
        #CA '_check_git_annex_version',
        # used by AnnexRepo.__init__() -- should be using `is_valid()`
        'is_valid_git',
        'is_valid_annex',
        '_is_direct_mode_from_config',
        #CA '_call_annex',
        #CA 'call_annex_items_',
    ]

    # intentionally limiting to just `path` as the only constructor argument
    def __new__(cls, path: Path):
        for attr in dir(cls):
            if not hasattr(LeanGitRepo, attr) \
                    and callable(getattr(cls, attr)) \
                    and attr not in LeanAnnexRepo._lean_attrs:
                setattr(cls, attr, _unsupported_method)

        obj = super(LegacyAnnexRepo, cls).__new__(cls)

        return obj


def _unsupported_method(self, *args, **kwargs):
    raise NotImplementedError('method unsupported by LeanAnnexRepo')



"""Add the method :meth:`url2transport_path` to RIA IO-abstraction classes

This patch adds the method :meth:`url2transport_path` to the IO-abstraction
classes: :class:`datalad.distributed.ora_remote.LocalIO`, and to the class
:class:`datalad.distributed.ora_remote.HTTPRemoteIO`.

This method is required by the patches that add Windows-client support to
RIA-code. It converts internally used abstract paths to concrete paths that are
platform- andIO-abstraction specific and on which IO-operations cam be
performed.
"""
from __future__ import annotations

import logging
from re import compile
from pathlib import (
    Path,
    PurePosixPath,
)

from datalad_next.consts import on_windows
from . import apply_patch


# The methods are patched into the ora_remote/ria_remote. Use the same logger.
lgr = logging.getLogger('datalad.customremotes.ria_remote')


drive_letter_matcher = compile('^/[A-Z]:')


def str2windows_path(url_path: PurePosixPath):
    path_str = str(url_path)
    match = drive_letter_matcher.match(path_str)
    if match:
        if path_str[3] == '/':
            return Path(*([f'{path_str[1]}:', '/'] + path_str[4:].split('/')))
        else:
            lgr.warning(f'Non-absolute Windows-path detected: {path_str}')
            return Path(*([f'{path_str[1]}:'] + path_str[3:].split('/')))
    else:
        return Path(path_str)


def local_io_url2transport_path(
        self,
        url_path: PurePosixPath
) -> Path | PurePosixPath:
    assert isinstance(url_path, PurePosixPath)
    if on_windows:
        return str2windows_path(url_path)
    else:
        return Path(url_path)


def http_remote_io_url2transport_path(
        self,
        url_path: PurePosixPath
) -> Path | PurePosixPath:
    assert isinstance(url_path, PurePosixPath)
    return url_path


# Add a `url2transport_path`-method to `ora_remote.LocalIO`
apply_patch(
    'datalad.distributed.ora_remote',
    'LocalIO',
    'url2transport_path',
    local_io_url2transport_path,
    expect_attr_present=False,
)


# Add a `url2transport_path`-method to `ora_remote.HTTPRemoteIO`
apply_patch(
    'datalad.distributed.ora_remote',
    'HTTPRemoteIO',
    'url2transport_path',
    http_remote_io_url2transport_path,
    expect_attr_present=False,
)



"""Report on the content of a Git repository worktree

The main functionality is provided by the :func:`iter_gitworktree()` function.
"""
from __future__ import annotations

from dataclasses import dataclass
from itertools import chain
import logging
from pathlib import (
    Path,
    PurePath,
    PurePosixPath,
)
from typing import (
    Iterator,
    Dict,
    Generator,
    Tuple,
)
from datasalad.itertools import (
    decode_bytes,
    itemize,
)

from datalad_next.runners import iter_git_subproc
from datasalad.gitpathspec import GitPathSpecs
from .utils import (
    FileSystemItem,
    FileSystemItemType,
)
from .gittree import (
    GitTreeItem,
    GitTreeItemType,
    _mode_type_map,
)


lgr = logging.getLogger('datalad.ext.next.iter_collections.gitworktree')


@dataclass
class GitWorktreeItem(GitTreeItem):
    name: PurePath


@dataclass
class GitWorktreeFileSystemItem(FileSystemItem):
    name: PurePath
    # gitsha is not the sha1 of the file content, but the output
    # of `git hash-object` which does something like
    # `printf "blob $(wc -c < "$file_name")\0$(cat "$file_name")" | sha1sum`
    gitsha: str | None = None
    gittype: GitTreeItemType | None = None

    @classmethod
    def from_worktreeitem(
        cls,
        basepath: Path,
        item: GitWorktreeItem,
        link_target: bool = True,
    ) -> "GitWorktreeFileSystemItem":
        """Create GitWorktreeFileSystemItem from corresponding GitWorktreeItem

        Parameters
        ----------
        basepath: Path
          Reference path to convert the ``GitWorktreeItem``'s path into a
          path on the file system.
        item: GitWorktreeItem
          Item to create matching ``GitWorktreeFileSystemItem`` for.
        link_target: bool
          Flag whether to read out a link-target for an item that is a symlink.
        """
        fsitem = GitWorktreeFileSystemItem.from_path(
            path=basepath / item.path,
            link_target=link_target,
        )
        fsitem.name = item.name
        fsitem.gitsha = item.gitsha
        fsitem.gittype = item.gittype
        return fsitem


lsfiles_untracked_args = {
    None:
    ('--stage', '--cached'),
    'all':
    ('--stage', '--cached', '--exclude-standard', '--others'),
    'whole-dir':
    ('--stage', '--cached', '--exclude-standard', '--others', '--directory'),
    'no-empty-dir':
    ('--stage', '--cached', '--exclude-standard',
     '--others', '--directory', '--no-empty-directory'),
    'only':
    ('--exclude-standard', '--others'),
    'only-whole-dir':
    ('--exclude-standard', '--others', '--directory'),
    'only-no-empty-dir':
    ('--exclude-standard',
     '--others', '--directory', '--no-empty-directory'),
}


def iter_gitworktree(
    path: Path,
    *,
    untracked: str | None = 'all',
    link_target: bool = False,
    fp: bool = False,
    recursive: str = 'repository',
    pathspecs: list[str] | GitPathSpecs | None = None,
) -> Generator[GitWorktreeItem | GitWorktreeFileSystemItem, None, None]:
    """Uses ``git ls-files`` to report on a work tree of a Git repository

    This iterator can be used to report on all tracked, and untracked content
    of a Git repository's work tree. This includes files that have been removed
    from the work tree (deleted), unless their removal has already been staged.

    For any tracked content, yielded items include type information and gitsha
    as last known to Git. This means that such reports reflect the last
    committed or staged content, not the state of a potential unstaged
    modification in the work tree.

    When no reporting of link targets or file-objects are requested, items of
    type :class:`GitWorktreeItem` are yielded, otherwise
    :class:`GitWorktreeFileSystemItem` instances. In both cases, ``gitsha`` and
    ``gittype`` properties are provided. Either of them being ``None``
    indicates untracked work tree content.

    .. note::
      The ``gitsha`` is not equivalent to a SHA1 hash of a file's content,
      but is the SHA-type blob identifier as reported and used by Git.

    Parameters
    ----------
    path: Path
      Path of a directory in a Git repository to report on. This directory
      need not be the root directory of the repository, but must be part of
      the repository's work tree.
    untracked: {'all', 'whole-dir', 'no-empty-dir', 'only', 'only-whole-dir', 'only-no-empty-dir'} or None, optional
      If not ``None``, also reports on untracked work tree content.
      ``all`` reports on any untracked file; ``whole-dir`` yields a single
      report for a directory that is entirely untracked, and not individual
      untracked files in it; ``no-empty-dir`` skips any reports on
      untracked empty directories. The modes starting with 'only' offer the
      same untracked content reporting styles, but only untracked and no
      tracked content is reported. For example, 'only' is the corresponding
      mode to 'all' with no tracked content being reported.
    link_target: bool, optional
      If ``True``, information matching a
      :class:`~datalad_next.iter_collections.utils.FileSystemItem`
      will be included for each yielded item, and the targets of
      any symlinks will be reported, too.
    fp: bool, optional
      If ``True``, information matching a
      :class:`~datalad_next.iter_collections.utils.FileSystemItem`
      will be included for each yielded item, but without a
      link target detection, unless ``link_target`` is given.
      Moreover, each file-type item includes a file-like object
      to access the file's content. This file handle will be closed
      automatically when the next item is yielded.
    recursive: {'submodules', 'repository', 'no'}, optional
      Behavior for recursion into subdirectories of ``path``. By default
      (``repository``), all directories within the repository are reported.
      This possibly includes untracked ones (see ``untracked``), but not
      directories within submodules. With ``submodules``, the full worktree
      is reported on with recursion into submodules. With ``no``,
      only direct children of ``path`` are reported on.
      For any worktree items in subdirectories of ``path`` only a single
      record for the containing immediate subdirectory ``path`` is yielded.
      For example, with 'path/subdir/file1' and 'path/subdir/file2' there
      will only be a single item with ``name='subdir'`` and
      ``type='directory'``.
    pathspecs: optional
      Patterns used to limit results to particular paths. Any pathspecs
      supported by Git can be used and are passed to the underlying ``git
      ls-files`` queries. Pathspecs are also supported for recursive reporting
      on submodules. In such a case, the results match those of individual
      queries with analog pathspecs on the respective submodules (Git itself
      does not support pathspecs for submodule-recursive operations).  For
      example, a ``submodule`` recursion with a pathspec ``*.jpg`` will yield
      reports on all JPG files in all submodules, even though a submodule path
      itself does not match ``*.jpg``.  On the other hand, a pathspec
      ``submoddir/*.jpg`` will only report on JPG files in the submodule at
      ``submoddir/``, but on all JPG files in that submodule.
      As of version 1.5, the pathspec support for submodule recursion is
      preliminary and results should be carefully investigated.

    Yields
    ------
    :class:`GitWorktreeItem` or :class:`GitWorktreeFileSystemItem`
      The ``name`` attribute of an item is a ``PurePath`` instance with
      the corresponding (relative) path, in platform conventions.
    """
    # we force-convert to Path to prevent delayed crashing when reading from
    # the file system. The docs already ask for that, but it is easy to
    # forget/ignore and leads to non-obvious errors. Running this once is
    # a cheap safety net
    # https://github.com/datalad/datalad-next/issues/551
    path = Path(path)
    _pathspecs = GitPathSpecs(pathspecs)

    processed_submodules: set[PurePath] = set()

    # the helper takes care of talking to Git and doing recursion
    for item in _iter_gitworktree(
        path=path,
        untracked=untracked,
        # the helper cannot do submodule recursion, we do this outside,
        # so limit here
        recursive='repository' if recursive == 'submodules' else recursive,
        pathspecs=_pathspecs,
    ):
        # exclude non-submodules, or a submodule that was found at
        # the root path -- which would indicate that the submodule
        # itself it not around, only its record in the parent
        if recursive == 'submodules' \
                and item.gittype == GitTreeItemType.submodule \
                and not item.name == PurePath('.'):
            # mark as processed immediately, independent of whether anything
            # need to be reported
            processed_submodules.add(item.name)
            yield from _yield_from_submodule(
                basepath=path,
                subm=item,
                untracked=untracked,
                recursive=recursive,
                link_target=link_target,
                fp=fp,
                pathspecs=_pathspecs,
            )
            # nothing else to do here, the iter_gitworktree() called
            # dealt with this submodule completely
            continue

        # here we take care of the file system related information,
        # reading out symlinks and opening files
        if link_target or fp:
            # convert to FileSystemItem and read out any symlinks
            try:
                item = GitWorktreeFileSystemItem.from_worktreeitem(
                    path,
                    item,
                    link_target=link_target,
                )
            except FileNotFoundError:
                pass
        # try opening the file
        # _get_fp_src() returns None of there is nothing to open
        fp_src = _get_fp_src(fp, path, item)
        if fp_src is None:
            # nothing to open
            yield item
        else:
            with fp_src.open('rb') as active_fp:
                item.fp = active_fp
                yield item

    # we may need to loop over the (remaining) submodules for two reasons:
    # - with pathspecs there is a chance that a given pathspec set did not
    #   match a submodule (directly) that could have content that matches a
    #   pathspec
    # - when we are looking for untracked content only, the code above
    #   (by definition) will not have found the submodules (because they are
    #   unconditionally tracked)
    if recursive == 'submodules' and (
        (untracked and untracked.startswith('only')) or _pathspecs
    ):
        for subm in iter_submodules(
            path=path,
            pathspecs=_pathspecs,
            match_containing=True,
        ):
            if subm.name in processed_submodules:
                # we dealt with that above already
                continue
            yield from _yield_from_submodule(
                basepath=path,
                subm=subm,
                untracked=untracked,
                recursive=recursive,
                link_target=link_target,
                fp=fp,
                pathspecs=_pathspecs,
            )


def _yield_from_submodule(
    basepath: Path,
    subm: GitTreeItem,
    untracked: str | None,
    recursive: str,
    link_target: bool,
    fp: bool,
    pathspecs: GitPathSpecs,
) -> Generator[GitWorktreeItem | GitWorktreeFileSystemItem, None, None]:
    # GitTreeItem.name is a str in POSIX notation. Convert to proper type
    # to get a meaningful path on all platforms
    subm_name = PurePosixPath(subm.name)
    subm_path = basepath / subm_name
    if not subm_path.exists():
        # no point in trying to list a submodule that is not around
        return
    subm_pathspecs = pathspecs
    if pathspecs:
        # recode pathspecs to match the submodule scope
        try:
            subm_pathspecs = pathspecs.for_subdir(subm_name)
        except ValueError:
            # not a single pathspec could be translated, there is
            # no chance for a match, we can stop here
            return
    for item in iter_gitworktree(
        path=subm_path,
        untracked=untracked,
        link_target=link_target,
        fp=fp,
        recursive=recursive,
        pathspecs=subm_pathspecs,
    ):
        # recode path/name
        item.name = subm.name / item.name
        # clear any possibly cached path value
        try:
            # if there is a cache, it is an instance value
            # with the cached_property name
            del item.path
        except AttributeError:
            pass
        yield item


def _iter_gitworktree(
    path: Path,
    *,
    untracked: str | None,
    recursive: str,
    pathspecs: GitPathSpecs,
) -> Generator[GitWorktreeItem, None, None]:
    """Internal helper for iter_gitworktree() tp support recursion"""

    # perform an implicit test of whether the `untracked` mode is known
    lsfiles_args = list(lsfiles_untracked_args[untracked])

    if pathspecs:
        lsfiles_args.extend(pathspecs.arglist())

    # helper to handle multi-stage reports by ls-files
    pending_item: tuple[None | PurePosixPath, None | Dict[str, str]] = (None, None)

    reported_dirs: set[PurePosixPath] = set()
    _single_dir = recursive == 'no'

    # we add a "fake" `None` record at the end to avoid a special
    # case for submitting the last pending item after the loop.
    # otherwise the context manager handling of the file pointer
    # would lead to lots of code duplication
    for line in chain(_git_ls_files(path, *lsfiles_args), [None]):
        # a bit ugly, but we need to account for the `None` record
        # that signals the final loop iteration
        ipath, lsfiles_props = _lsfiles_line2props(line) \
            if line is not None else (None, None)

        # yield any pending item, if the current record is not an
        # addendum of it
        if ipath is None or pending_item[0] not in (None, ipath):
            if ipath is None and pending_item[0] is None:
                return
            assert pending_item[0] is not None
            # this is the last point where we can still withhold a report.
            # it is also the point where we can do this with minimal
            # impact on the rest of the logic.
            # so act on recursion setup now
            pending_item_path_parts = pending_item[0].parts
            if _single_dir and len(pending_item_path_parts) > 1:
                # this path is pointing inside a subdirectory of the
                # base directory -> ignore
                # we do reset pending_item here, although this would also
                # happen below -- it decomplexifies the conditionals
                dir_path = PurePosixPath(pending_item_path_parts[0])
                if dir_path in reported_dirs:
                    # we only yield each containing dir once, and only once
                    pending_item = (ipath, lsfiles_props)
                    continue
                item = _get_item(
                    path,
                    # we know all props already
                    ipath=dir_path,
                    type=GitTreeItemType.directory,
                    gitsha=None,
                )
                yield item
                reported_dirs.add(dir_path)
                pending_item = (ipath, lsfiles_props)
                continue

            assert pending_item[0] is not None
            # report on a pending item, this is not a "higher-stage"
            # report by ls-files
            item = _get_item(
                path,
                pending_item[0],
                pending_item[1]['mode'] if pending_item[1] else None,
                pending_item[1]['gitsha'] if pending_item[1] else None,
            )
            yield item

        if ipath is None:
            # this is the trailing `None` record. we are done here
            break

        if lsfiles_props is None:
            # when no properties were produced, this is a
            # category "other" report (i.e., untracked content)
            # the path is always relative-POSIX
            pending_item = (ipath, None)
        else:
            pending_item = (ipath, lsfiles_props)
        # do not yield immediately, wait for a possible higher-stage
        # report in the next loop iteration


def iter_submodules(
    path: Path,
    *,
    pathspecs: list[str] | GitPathSpecs | None = None,
    match_containing: bool = False,
) -> Generator[GitTreeItem, None, None]:
    """Given a path, report all submodules of a repository worktree underneath

    With ``match_containing`` set to the default ``False``, this is merely a
    convenience wrapper around ``iter_gitworktree()`` that selectively reports
    on submodules. With ``match_containing=True`` and ``pathspecs`` given, the
    yielded items corresponding to submodules where the given ``pathsspecs``
    *could* match content. This includes submodules that are not available
    locally, because no actual matching of pathspecs to submodule content is
    performed -- only an evaluation of the submodule item itself.
    """
    _pathspecs = GitPathSpecs(pathspecs)
    if not _pathspecs:
        # force flag to be sensible to simplify internal logic
        match_containing = False

    for item in iter_gitworktree(
        path,
        untracked=None,
        link_target=False,
        fp=False,
        recursive='repository',
        # if we want to match submodules that contain pathspecs matches
        # we cannot give the job to Git, it won't report anything,
        # but we need to match manually below
        pathspecs=None if match_containing else _pathspecs,
    ):
        # exclude non-submodules, or a submodule that was found at
        # the root path -- which would indicate that the submodule
        # itself it not around, only its record in the parent
        if item.gittype != GitTreeItemType.submodule \
                or item.name == PurePath('.'):
            continue

        if not match_containing:
            yield item
            continue

        assert pathspecs is not None
        # does any pathspec match the "inside" of the current submodule's
        # path
        if _pathspecs.any_match_subdir(PurePosixPath(item.name)):
            yield item
            continue

        # no match
        continue


def _get_item(
    basepath: Path,
    ipath: PurePosixPath,
    type: str | GitTreeItemType | None = None,
    gitsha: str | None = None,
) -> GitWorktreeItem:
    if isinstance(type, str):
        type = _mode_type_map[type]
    item = GitWorktreeItem(name=ipath)
    if type is not None:
        item.gittype = type
    if gitsha is not None:
        item.gitsha = gitsha
    # make sure the name/id is the path relative to the basepath
    item.name = PurePath(ipath)
    return item


def _lsfiles_line2props(
    line: str
) -> Tuple[PurePosixPath, Dict[str, str] | None]:
    items = line.split('\t', maxsplit=1)
    # check if we cannot possibly have a 'staged' report with mode and gitsha
    if len(items) < 2:
        # not known to Git, but Git always reports POSIX
        path = PurePosixPath(line)
        # early exist, we have nothing but the path (untracked)
        return path, None

    props = items[0].split(' ')
    if len(props) != 3:
        # not known to Git, but Git always reports POSIX
        path = PurePosixPath(line)
        # early exist, we have nothing but the path (untracked)
        return path, None

    # again Git reports always in POSIX
    path = PurePosixPath(items[1])
    return path, dict(
        gitsha=props[1],
        mode=props[0],
    )


def _git_ls_files(path: Path, *args) -> Iterator[str]:
    with iter_git_subproc(
            [
                'ls-files',
                # we rely on zero-byte splitting below
                '-z',
                # otherwise take whatever is coming in
                *args,
            ],
            cwd=path,
    ) as r:
        yield from itemize(
            decode_bytes(r, backslash_replace=True),
            sep='\0',
            keep_ends=False,
        )


def _get_fp_src(
    fp: bool,
    basepath: Path,
    item: GitWorktreeItem | GitWorktreeFileSystemItem,
) -> Path | None:
    if not fp or isinstance(item, GitWorktreeItem):
        # no file pointer request, we are done
        return None

    # if we get here, this is about file pointers...
    fp_src = None
    if item.type in (FileSystemItemType.file,
                     FileSystemItemType.symlink):
        fp_src = item.name
    if fp_src is None:
        # nothing to open
        return None

    fp_src_fullpath = basepath / fp_src
    if not fp_src_fullpath.exists():
        # nothing there to open (would resolve through a symlink)
        return None

    return fp_src_fullpath



from __future__ import annotations

from importlib import import_module
import logging
from typing import Any

lgr = logging.getLogger('datalad.ext.next.patches')


def apply_patch(
        modname: str,
        objname: str | None,
        attrname: str,
        patch: Any,
        msg: str | None = None,
        expect_attr_present=True,
):
    """
    Monkey patch helper

    Parameters
    ----------
    modname: str
      Importable name of the module with the patch target.
    objname: str or None
      If `None`, patch will target an attribute of the module,
      otherwise an object name within the module can be specified.
    attrname: str
      Name of the attribute to replace with the patch, either
      in the module or the given object (see ``objname``)
    patch:
      The identified attribute will be replaced with this object.
    msg: str or None
      If given, a debug-level log message with this text will be
      emitted, otherwise a default message is generated.
    expect_attr_present: bool
      If True (default) an exception is raised when the target
      attribute is not found.

    Returns
    -------
    object
      The original, unpatched attribute -- if ``expect_attr_present``
      is enabled, or ``None`` otherwise.

    Raises
    ------
    ImportError
      When the target module cannot be imported
    AttributedError
      When the target object is not found, or the target attribute is
      not found.
    """
    orig_attr = None

    msg = msg or f'Apply patch to {modname}.{attrname}'
    lgr.debug(msg)

    # we want to fail on ImportError
    mod = import_module(modname, package='datalad')
    if objname:
        # we want to fail on a missing object
        obj = getattr(mod, objname)
    else:
        # the target is the module itself
        obj = mod

    if expect_attr_present:
        # we want to fail on a missing attribute/object
        orig_attr = getattr(obj, attrname)

    setattr(obj, attrname, patch)

    return orig_attr



from datalad_next.tests import skip_wo_symlink_capability

from ..utils import FileSystemItem


def test_FileSystemItem(tmp_path):
    testfile = tmp_path / 'file1.txt'
    testfile_content = 'content'
    testfile.write_text(testfile_content)

    item = FileSystemItem.from_path(testfile)
    assert item.size == len(testfile_content)
    assert item.link_target is None


@skip_wo_symlink_capability
def test_FileSystemItem_linktarget(tmp_path):
    testfile = tmp_path / 'file1.txt'
    testfile_content = 'short'
    testfile.write_text(testfile_content)
    testlink = tmp_path / 'link'
    testlink.symlink_to(testfile)

    item = FileSystemItem.from_path(testlink)
    assert testfile.samefile(item.link_target)
    # size of the link file does not anyhow propagate the size of the
    # link target
    assert item.size != len(testfile_content)

    # we can disable link resolution
    item = FileSystemItem.from_path(testlink, link_target=False)
    assert item.link_target is None



"""ria-, ora-, ria_utils-, and clone-tests for patched ria/ora code

This are patched tests that work with the patched ria/ora code.

The ria-utils-patches use an abstract path representation for RIA-store elements.
This patch adapts the tests that use `ria_utils.create_store` and
`ria_utils.create_ds_in_store` to these modifications.
"""
from __future__ import annotations

import logging
import os.path as op
import random
import shutil
import stat
import string
from pathlib import (
    Path,
    PurePosixPath,
)
from urllib.parse import urlparse
from urllib.request import pathname2url

from datalad.api import (
    Dataset,
    clone,
    create_sibling_ria,
)
from datalad.cmd import (
    WitlessRunner as Runner,
    NoCapture,
)
from datalad.customremotes.ria_utils import (
    UnknownLayoutVersion,
    create_ds_in_store,
    create_store,
    get_layout_locations,
)
from datalad.distributed.ora_remote import (
    LocalIO,
    SSHRemoteIO,
)
from datalad.distributed.tests.ria_utils import (
    common_init_opts,
    get_all_files,
    populate_dataset,
)
from datalad.support.exceptions import (
    CommandError,
    IncompleteResultsError,
)
from datalad.support.network import get_local_file_url
from datalad.tests.utils_pytest import (
    DEFAULT_REMOTE,
    SkipTest,
    assert_equal,
    assert_false,
    assert_not_in,
    assert_not_is_instance,
    assert_repo_status,
    assert_result_count,
    assert_status,
    assert_true,
    create_tree,
    has_symlink_capability,
    known_failure_windows,
    ok_,
    rmtree,
    serve_path_via_http,
    skip_if_adjusted_branch,
    skip_if_root,
    skip_ssh,
    slow,
    turtle,
    with_tempfile,
)

from datalad_next.tests import (
    assert_in,
    assert_raises,
    eq_,
    swallow_logs,
)


def _local_path2pure_posix_path(path: Path | str):
    return PurePosixPath(pathname2url(str(path)))


def _random_name(prefix: str = '') -> str:
    return prefix + ''.join(
        random.choices(string.ascii_letters + string.digits, k=8)
    )


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
@with_tempfile
def _postclonetest_prepare(lcl, storepath, storepath2, link):

    from datalad.customremotes.ria_utils import (
        create_ds_in_store,
        create_store,
        get_layout_locations,
    )
    from datalad.distributed.ora_remote import LocalIO

    create_tree(lcl,
                tree={
                        'ds': {
                            'test.txt': 'some',
                            'subdir': {
                                'subds': {'testsub.txt': 'somemore'},
                                'subgit': {'testgit.txt': 'even more'}
                            },
                        },
                      })

    lcl = Path(lcl)
    storepath = Path(storepath)
    storepath2 = Path(storepath2)
    # PATCH: introduce `ppp_storepath` and `ppp_storepath2` and use them instead
    # of `storepath` and `storepath2`.
    ppp_storepath = _local_path2pure_posix_path(storepath)
    ppp_storepath2 = _local_path2pure_posix_path(storepath2)
    link = Path(link)
    link.symlink_to(storepath)

    # create a local dataset with a subdataset
    subds = Dataset(lcl / 'ds' / 'subdir' / 'subds').create(force=True)
    subds.save()
    # add a plain git dataset as well
    subgit = Dataset(lcl / 'ds' / 'subdir' / 'subgit').create(force=True,
                                                              annex=False)
    subgit.save()
    ds = Dataset(lcl / 'ds').create(force=True)
    ds.save(version_tag='original')
    assert_repo_status(ds.path)

    io = LocalIO()

    # Have a second store with valid ORA remote. This should not interfere with
    # reconfiguration of the first one, when that second store is not the one we
    # clone from. However, don't push data into it for easier get-based testing
    # later on.
    # Doing this first, so datasets in "first"/primary store know about this.
    create_store(io, ppp_storepath2, '1')
    url2 = "ria+{}".format(get_local_file_url(str(storepath2)))
    for d in (ds, subds, subgit):
        create_ds_in_store(io, ppp_storepath2, d.id, '2', '1')
        d.create_sibling_ria(url2, "anotherstore", new_store_ok=True)
        d.push('.', to='anotherstore', data='nothing')
        store2_loc, _, _ = get_layout_locations(1, storepath2, d.id)
        Runner(cwd=str(store2_loc)).run(['git', 'update-server-info'])

    # Now the store to clone from:
    create_store(io, ppp_storepath, '1')

    # URL to use for upload. Point is, that this should be invalid for the clone
    # so that autoenable would fail. Therefore let it be based on a to be
    # deleted symlink
    upl_url = "ria+{}".format(get_local_file_url(str(link)))

    for d in (ds, subds, subgit):

        # TODO: create-sibling-ria required for config! => adapt to RF'd
        #       creation (missed on rebase?)
        create_ds_in_store(io, ppp_storepath, d.id, '2', '1')
        d.create_sibling_ria(upl_url, "store", new_store_ok=True)

        if d is not subgit:
            # Now, simulate the problem by reconfiguring the special remote to
            # not be autoenabled.
            # Note, however, that the actual intention is a URL, that isn't
            # valid from the point of view of the clone (doesn't resolve, no
            # credentials, etc.) and therefore autoenabling on git-annex-init
            # when datalad-cloning would fail to succeed.
            Runner(cwd=d.path).run(['git', 'annex', 'enableremote',
                                    'store-storage',
                                    'autoenable=false'])
        d.push('.', to='store')
        store_loc, _, _ = get_layout_locations(1, storepath, d.id)
        Runner(cwd=str(store_loc)).run(['git', 'update-server-info'])

    link.unlink()
    # We should now have a store with datasets that have an autoenabled ORA
    # remote relying on an inaccessible URL.
    # datalad-clone is supposed to reconfigure based on the URL we cloned from.
    # Test this feature for cloning via HTTP, SSH and FILE URLs.

    return ds.id


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732 and patched
# and refactored
def test_ria_postclone_noannex(existing_dataset, tmp_path):
    # Test for gh-5186: Cloning from local FS, shouldn't lead to annex
    # initializing origin.
    from datalad.customremotes.ria_utils import (
        create_ds_in_store,
        create_store,
        get_layout_locations,
    )
    from datalad.distributed.ora_remote import LocalIO

    ds = existing_dataset
    some_file = ds.pathobj / 'a_file.txt'
    some_file.write_text('irrelevant')

    ria_store_path = tmp_path / 'ria_store'
    url_ria_store_path = _local_path2pure_posix_path(ria_store_path)

    io = LocalIO()
    create_store(io, url_ria_store_path, '1')
    lcl_url = "ria+{}".format(get_local_file_url(str(ria_store_path)))
    create_ds_in_store(io, url_ria_store_path, ds.id, '2', '1')
    ds.create_sibling_ria(lcl_url, "store", new_store_ok=True)
    ds.push('.', to='store')

    # Remove annex-tree from store to check that clone doesn't cause annex to
    # recreate it.
    store_loc = get_layout_locations(1, ria_store_path, ds.id)[0]
    annex = store_loc / 'annex'
    rmtree(str(annex))
    assert_false(annex.exists())

    clone(
        f"ria+{get_local_file_url(str(ria_store_path), compatibility='git')}#{ds.id}",
        tmp_path / 'cloned_ds'
    )
    assert_false(annex.exists())


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
def _test_setup_store(io, store_url, local_store_path):

    store_url_path = PurePosixPath(urlparse(store_url).path)

    # invalid version raises:
    assert_raises(
        UnknownLayoutVersion,
        create_store, io, store_url_path, '2')

    # non-existing path should work:
    create_store(io, store_url_path, '1')

    version_file = local_store_path / 'ria-layout-version'
    error_logs = local_store_path / 'error_logs'
    assert_true(version_file.exists())
    assert_true(error_logs.exists())
    assert_true(error_logs.is_dir())
    assert_equal([f for f in error_logs.iterdir()], [])

    # empty target directory should work as well:
    rmtree(str(local_store_path))
    local_store_path.mkdir(exist_ok=False)
    create_store(io, store_url_path, '1')
    assert_true(version_file.exists())
    assert_true(error_logs.exists())
    assert_true(error_logs.is_dir())
    assert_equal([f for f in error_logs.iterdir()], [])

    # re-execution also fine:
    create_store(io, store_url_path, '1')

    # but version conflict with existing target isn't:
    version_file.write_text("2|unknownflags\n")
    assert_raises(ValueError, create_store, io, store_url_path, '1')
    # TODO: check output reporting conflicting version "2"


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
def _test_setup_ds_in_store(io, store_url, local_store_path):

    store_url_path = PurePosixPath(urlparse(store_url).path)

    # ATM create_ds_in_store doesn't care what kind of ID is provided
    dsid = "abc123456"

    ds_path = local_store_path / dsid[:3] / dsid[3:]  # store layout version 1
    version_file = ds_path / 'ria-layout-version'
    archives = ds_path / 'archives'
    objects = ds_path / 'annex' / 'objects'
    git_config = ds_path / 'config'

    # invalid store version:
    assert_raises(
        UnknownLayoutVersion,
        create_ds_in_store, io, store_url_path, dsid, '1', 'abc'
    )

    # invalid obj version:
    assert_raises(
        UnknownLayoutVersion,
        create_ds_in_store, io, store_url_path, dsid, 'abc', '1'
    )

    # version 1
    create_store(io, store_url_path, '1')
    create_ds_in_store(io, store_url_path, dsid, '1', '1')
    for p in [ds_path, archives, objects]:
        assert_true(p.is_dir(), msg="Not a directory: %s" % str(p))
    for p in [version_file]:
        assert_true(p.is_file(), msg="Not a file: %s" % str(p))
    assert_equal(version_file.read_text().strip(), "1")

    # conflicting version exists at target:
    assert_raises(
        ValueError,
        create_ds_in_store, io, store_url_path, dsid, '2', '1'
    )

    # version 2
    # Note: The only difference between version 1 and 2 are supposed to be the
    #       key paths (dirhashlower vs mixed), which has nothing to do with
    #       setup routine.
    rmtree(local_store_path)
    create_store(io, store_url_path, '1')
    create_ds_in_store(io, store_url_path, dsid, '2', '1')
    for p in [ds_path, archives, objects]:
        assert_true(p.is_dir(), msg="Not a directory: %s" % str(p))
    for p in [version_file]:
        assert_true(p.is_file(), msg="Not a file: %s" % str(p))
    assert_equal(version_file.read_text().strip(), "2")


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
@with_tempfile(mkdir=True)
@serve_path_via_http
@with_tempfile
def test_initremote(store_path=None, store_url=None, ds_path=None):
    ds = Dataset(ds_path).create()
    store_path = Path(store_path)
    # PATCH: introduce `ppp_store_path` and use it instead of `store_path`
    ppp_store_path = _local_path2pure_posix_path(store_path)
    url = "ria+" + store_url
    init_opts = common_init_opts + ['url={}'.format(url)]

    # fail when there's no RIA store at the destination
    assert_raises(CommandError, ds.repo.init_remote, 'ora-remote',
                  options=init_opts)
    # Doesn't actually create a remote if it fails
    assert_not_in('ora-remote',
                  [cfg['name']
                   for uuid, cfg in ds.repo.get_special_remotes().items()]
                  )

    # now make it a store
    io = LocalIO()
    create_store(io, ppp_store_path, '1')
    create_ds_in_store(io, ppp_store_path, ds.id, '2', '1')

    # fails on non-RIA URL
    assert_raises(CommandError, ds.repo.init_remote, 'ora-remote',
                  options=common_init_opts + ['url={}'
                                              ''.format(store_path.as_uri())]
                  )
    # Doesn't actually create a remote if it fails
    assert_not_in('ora-remote',
                  [cfg['name']
                   for uuid, cfg in ds.repo.get_special_remotes().items()]
                  )

    ds.repo.init_remote('ora-remote', options=init_opts)
    assert_in('ora-remote',
              [cfg['name']
               for uuid, cfg in ds.repo.get_special_remotes().items()]
              )
    assert_repo_status(ds.path)
    # git-annex:remote.log should have:
    #   - url
    #   - common_init_opts
    #   - archive_id (which equals ds id)
    remote_log = ds.repo.call_git(['cat-file', 'blob', 'git-annex:remote.log'],
                                  read_only=True)
    assert_in("url={}".format(url), remote_log)
    [assert_in(c, remote_log) for c in common_init_opts]
    assert_in("archive-id={}".format(ds.id), remote_log)


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
# TODO: on crippled FS copytree to populate store doesn't seem to work.
#       Or may be it's just the serving via HTTP that doesn't work.
#       Either way, after copytree and fsck, whereis doesn't report
#       the store as an available source.
@skip_if_adjusted_branch
@known_failure_windows  # see gh-4469
@with_tempfile(mkdir=True)
@serve_path_via_http
@with_tempfile
def test_read_access(store_path=None, store_url=None, ds_path=None):

    ds = Dataset(ds_path).create()
    populate_dataset(ds)

    files = [Path('one.txt'), Path('subdir') / 'two']
    store_path = Path(store_path)
    # PATCH: introduce `ppp_store_path` and use it instead of `store_path`
    ppp_store_path = _local_path2pure_posix_path(store_path)
    url = "ria+" + store_url
    init_opts = common_init_opts + ['url={}'.format(url)]

    io = LocalIO()
    create_store(io, ppp_store_path, '1')
    create_ds_in_store(io, ppp_store_path, ds.id, '2', '1')
    ds.repo.init_remote('ora-remote', options=init_opts)
    fsck_results = ds.repo.fsck(remote='ora-remote', fast=True)
    # Note: Failures in the special remote will show up as a success=False
    # result for fsck -> the call itself would not fail.
    for r in fsck_results:
        if "note" in r:
            # we could simply assert "note" to not be in r, but we want proper
            # error reporting - content of note, not just its unexpected
            # existence.
            assert_equal(r["success"], "true",
                         msg="git-annex-fsck failed with ORA over HTTP: %s" % r)
        assert_equal(r["error-messages"], [])
    store_uuid = ds.siblings(name='ora-remote',
                             return_type='item-or-list',
                             result_renderer='disabled')['annex-uuid']
    here_uuid = ds.siblings(name='here',
                            return_type='item-or-list',
                            result_renderer='disabled')['annex-uuid']

    # nothing in store yet:
    for f in files:
        known_sources = ds.repo.whereis(str(f))
        assert_in(here_uuid, known_sources)
        assert_not_in(store_uuid, known_sources)

    annex_obj_target = str(store_path / ds.id[:3] / ds.id[3:]
                           / 'annex' / 'objects')
    shutil.rmtree(annex_obj_target)
    shutil.copytree(src=str(ds.repo.dot_git / 'annex' / 'objects'),
                    dst=annex_obj_target)

    ds.repo.fsck(remote='ora-remote', fast=True)
    # all in store now:
    for f in files:
        known_sources = ds.repo.whereis(str(f))
        assert_in(here_uuid, known_sources)
        assert_in(store_uuid, known_sources)

    ds.drop('.')
    res = ds.get('.')
    assert_equal(len(res), 4)
    assert_result_count(res, 4, status='ok', type='file', action='get',
                        message="from ora-remote...")

    # try whether the reported access URL is correct
    one_url = ds.repo.whereis('one.txt', output='full'
        )[store_uuid]['urls'].pop()
    assert_status('ok', ds.download_url(urls=[one_url], path=str(ds.pathobj / 'dummy')))


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
def _test_initremote_basic(io, store_url, local_store_path, ds, link, overwrite = None):

    populate_dataset(ds)

    store_url_path = overwrite or PurePosixPath(urlparse(store_url).path)
    link = Path(link)

    init_opts = common_init_opts + [f'url=ria+{store_url}']

    # fails on non-existing storage location
    assert_raises(
        CommandError,
        ds.repo.init_remote, 'ria-remote', options=init_opts
    )
    # Doesn't actually create a remote if it fails
    assert_not_in(
        'ria-remote',
        [
            cfg['name']
            for uuid, cfg in ds.repo.get_special_remotes().items()
        ]
    )

    # fails on non-RIA URL
    assert_raises(
        CommandError,
        ds.repo.init_remote, 'ria-remote',
        options=common_init_opts + [f'url={store_url}']
    )

    # Doesn't actually create a remote if it fails
    assert_not_in(
        'ria-remote',
        [
            cfg['name']
            for uuid, cfg in ds.repo.get_special_remotes().items()
        ]
    )

    # set up store:
    create_store(io, store_url_path, '1')
    # still fails, since ds isn't setup in the store
    assert_raises(
        CommandError,
  ds.repo.init_remote, 'ria-remote', options=init_opts
    )
    # Doesn't actually create a remote if it fails
    assert_not_in(
        'ria-remote',
        [
            cfg['name']
            for uuid, cfg in ds.repo.get_special_remotes().items()
        ]
    )
    # set up the dataset as well
    create_ds_in_store(io, store_url_path, ds.id, '2', '1')
    # now should work
    ds.repo.init_remote('ria-remote', options=init_opts)
    assert_in(
        'ria-remote',
        [
            cfg['name']
            for uuid, cfg in ds.repo.get_special_remotes().items()
        ]
    )
    assert_repo_status(ds.path)
    # git-annex:remote.log should have:
    #   - url
    #   - common_init_opts
    #   - archive_id (which equals ds id)
    remote_log = ds.repo.call_git(
        ['cat-file', 'blob', 'git-annex:remote.log'], read_only=True
    )
    assert_in(f'url=ria+{store_url}', remote_log)
    [assert_in(c, remote_log) for c in common_init_opts]
    assert_in("archive-id={}".format(ds.id), remote_log)

    # re-configure with invalid URL should fail:
    assert_raises(
        CommandError,
        ds.repo.call_annex,
        ['enableremote', 'ria-remote'] + common_init_opts + [
            'url=ria+file:///non-existing'
        ]
    )
    # but re-configure with valid URL should work
    if has_symlink_capability():
        link.symlink_to(local_store_path)
        new_url = 'ria+{}'.format(link.as_uri())
        ds.repo.call_annex(
            ['enableremote', 'ria-remote'] + common_init_opts + [
                'url={}'.format(new_url)
            ]
        )
        # git-annex:remote.log should have:
        #   - url
        #   - common_init_opts
        #   - archive_id (which equals ds id)
        remote_log = ds.repo.call_git(
            ['cat-file', 'blob', 'git-annex:remote.log'], read_only=True
        )
        assert_in("url={}".format(new_url), remote_log)
        [assert_in(c, remote_log) for c in common_init_opts]
        assert_in("archive-id={}".format(ds.id), remote_log)

    # we can deal with --sameas, which leads to a special remote not having a
    # 'name' property, but only a 'sameas-name'. See gh-4259
    try:
        ds.repo.init_remote(
            'ora2',
            options=init_opts + ['--sameas', 'ria-remote']
        )
    except CommandError as e:
        if 'Invalid option `--sameas' in e.stderr:
            # annex too old - doesn't know --sameas
            pass
        else:
            raise
    # TODO: - check output of failures to verify it's failing the right way
    #       - might require to run initremote directly to get the output


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
@known_failure_windows  # see gh-4469
def _test_remote_layout(io,
                        store_url,
                        archive_store_url,
                        local_store_path,
                        local_archive_store_path,
                        ds
                        ):

    populate_dataset(ds)
    assert_repo_status(ds.path)

    store_url_path = PurePosixPath(urlparse(store_url).path)
    archive_store_url_path = PurePosixPath(urlparse(archive_store_url).path)

    # set up store:
    create_store(io, store_url_path, '1')

    # TODO: Re-establish test for version 1
    # version 2: dirhash
    create_ds_in_store(io, store_url_path, ds.id, '2', '1')

    # add special remote
    init_opts = common_init_opts + [f'url=ria+{store_url}']
    ds.repo.init_remote('store', options=init_opts)

    # copy files into the RIA store
    ds.push('.', to='store')

    # we should see the exact same annex object tree
    dsgit_dir, archive_dir, dsobj_dir = get_layout_locations(
        1, local_store_path, ds.id
    )
    store_objects = get_all_files(dsobj_dir)
    local_objects = get_all_files(ds.pathobj / '.git' / 'annex' / 'objects')
    assert_equal(len(store_objects), 4)

    if not ds.repo.is_managed_branch():
        # with managed branches the local repo uses hashdirlower instead
        # TODO: However, with dataset layout version 1 this should therefore
        #       work on adjusted branch the same way
        # TODO: Wonder whether export-archive-ora should account for that and
        #       rehash according to target layout.
        assert_equal(sorted([p for p in store_objects]),
                     sorted([p for p in local_objects])
                     )

        if not io.get_7z():
            raise SkipTest("No 7z available in RIA store")

        # we can simply pack up the content of the remote into a
        # 7z archive and place it in the right location to get a functional
        # archive remote

        create_store(io, archive_store_url_path, '1')
        create_ds_in_store(io, archive_store_url_path, ds.id, '2', '1')

        whereis = ds.repo.whereis('one.txt')
        dsgit_dir, archive_dir, dsobj_dir = get_layout_locations(
            1, local_archive_store_path, ds.id)
        ds.export_archive_ora(archive_dir / 'archive.7z')
        init_opts = common_init_opts + [f'url=ria+{archive_store_url}']
        ds.repo.init_remote('archive', options=init_opts)
        # now fsck the new remote to get the new special remote indexed
        ds.repo.fsck(remote='archive', fast=True)
        assert_equal(len(ds.repo.whereis('one.txt')), len(whereis) + 1)
        # test creating an archive with filters on files
        ds.export_archive_ora(archive_dir / 'archive2.7z', annex_wanted='(include=*.txt)')
        # test with wanted expression of a specific remote
        ds.repo.set_preferred_content("wanted", "include=subdir/*", remote="store")
        ds.export_archive_ora(archive_dir / 'archive3.7z', remote="store")
        # test with the current sha
        ds.export_archive_ora(
            archive_dir / 'archive4.7z',
            froms=ds.repo.get_revisions()[1],
            )


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
def _test_version_check(io, store_url, local_store_path, ds):

    store_url_path = PurePosixPath(urlparse(store_url).path)

    populate_dataset(ds)
    assert_repo_status(ds.path)

    create_store(io, store_url_path, '1')

    # TODO: Re-establish test for version 1
    # version 2: dirhash
    create_ds_in_store(io, store_url_path, ds.id, '2', '1')

    # add special remote
    init_opts = common_init_opts + [f'url={"ria+" + store_url}']
    ds.repo.init_remote('store', options=init_opts)
    ds.push('.', to='store')

    # check version files
    remote_ds_tree_version_file = local_store_path / 'ria-layout-version'
    dsgit_dir, archive_dir, dsobj_dir = get_layout_locations(
        1,
        local_store_path,
        ds.id
    )
    remote_obj_tree_version_file = dsgit_dir / 'ria-layout-version'

    assert_true(remote_ds_tree_version_file.exists())
    assert_true(remote_obj_tree_version_file.exists())

    with open(str(remote_ds_tree_version_file), 'r') as f:
        assert_equal(f.read().strip(), '1')
    with open(str(remote_obj_tree_version_file), 'r') as f:
        assert_equal(f.read().strip(), '2')

    # Accessing the remote should not yield any output regarding versioning,
    # since it's the "correct" version. Note that "fsck" is an arbitrary choice.
    # We need just something to talk to the special remote.
    with swallow_logs(new_level=logging.INFO, name='datalad.annex') as cml:
        ds.repo.fsck(remote='store', fast=True)
        # TODO: For some reason didn't get cml.assert_logged to assert
        #       "nothing was logged"
        assert not cml.out

    # Now fake-change the version
    with open(str(remote_obj_tree_version_file), 'w') as f:
        f.write('X\n')

    # Now we should see a message about it
    with swallow_logs(new_level=logging.INFO, name='datalad.annex') as cml:
        ds.repo.fsck(remote='store', fast=True)
        cml.assert_logged(level="INFO",
                          msg="Remote object tree reports version X",
                          regex=False)

    # reading still works:
    ds.drop('.')
    assert_status('ok', ds.get('.'))

    # but writing doesn't:
    with open(str(Path(ds.path) / 'new_file'), 'w') as f:
        f.write("arbitrary addition")
    ds.save(message="Add a new_file")

    with assert_raises((CommandError, IncompleteResultsError)):
        ds.push('new_file', to='store')

    # However, we can force it by configuration
    ds.config.add("annex.ora-remote.store.force-write", "true", scope='local')
    ds.push('new_file', to='store')


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
# git-annex-testremote is way too slow on crippled FS.
# Use is_managed_branch() as a proxy and skip only here
# instead of in a decorator
@skip_if_adjusted_branch
@known_failure_windows  # see gh-4469
@with_tempfile
@with_tempfile
def _test_gitannex(host, store, dspath):
    dspath = Path(dspath)
    store = Path(store)
    # PATCH: introduce `ppp_store` and use it instead of `store`
    ppp_store = _local_path2pure_posix_path(store)

    ds = Dataset(dspath).create()

    populate_dataset(ds)
    assert_repo_status(ds.path)

    # set up store:
    io = SSHRemoteIO(host) if host else LocalIO()
    if host:
        store_url = "ria+{host}{path}".format(
            host=host[:-1],
            path=store
        )
    else:
        store_url = "ria+{}".format(store.as_uri())

    create_store(io, ppp_store, '1')

    # TODO: Re-establish test for version 1
    # version 2: dirhash
    create_ds_in_store(io, ppp_store, ds.id, '2', '1')

    # add special remote
    init_opts = common_init_opts + ['url={}'.format(store_url)]
    ds.repo.init_remote('store', options=init_opts)

    from datalad.support.external_versions import external_versions
    if '8.20200330' < external_versions['cmd:annex'] < '8.20200624':
        # https://git-annex.branchable.com/bugs/testremote_breeds_way_too_many_instances_of_the_externals_remote/?updated
        raise SkipTest(
            "git-annex might lead to overwhelming number of external "
            "special remote instances")

    # run git-annex-testremote
    # note, that we don't want to capture output. If something goes wrong we
    # want to see it in test build's output log.
    ds.repo._call_annex(['testremote', 'store'], protocol=NoCapture)


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
@known_failure_windows
@with_tempfile
@with_tempfile
@with_tempfile
def test_push_url(storepath=None, dspath=None, blockfile=None):

    dspath = Path(dspath)
    store = Path(storepath)
    # PATCH: introduce `ppp_store` and use it instead of `store`
    ppp_store = _local_path2pure_posix_path(store)
    blockfile = Path(blockfile)
    blockfile.touch()

    ds = Dataset(dspath).create()
    populate_dataset(ds)
    assert_repo_status(ds.path)
    repo = ds.repo

    # set up store:
    io = LocalIO()
    store_url = "ria+{}".format(ppp_store.as_uri())
    create_store(io, ppp_store, '1')
    create_ds_in_store(io, ppp_store, ds.id, '2', '1')

    # initremote fails with invalid url (not a ria+ URL):
    invalid_url = (store.parent / "non-existent").as_uri()
    init_opts = common_init_opts + ['url={}'.format(store_url),
                                    'push-url={}'.format(invalid_url)]
    assert_raises(CommandError, ds.repo.init_remote, 'store', options=init_opts)

    # initremote succeeds with valid but inaccessible URL (pointing to a file
    # instead of a store):
    block_url = "ria+" + blockfile.as_uri()
    init_opts = common_init_opts + ['url={}'.format(store_url),
                                    'push-url={}'.format(block_url)]
    repo.init_remote('store', options=init_opts)

    store_uuid = ds.siblings(name='store',
                             return_type='item-or-list')['annex-uuid']
    here_uuid = ds.siblings(name='here',
                            return_type='item-or-list')['annex-uuid']

    # but a push will fail:
    assert_raises(CommandError, ds.repo.call_annex,
                  ['copy', 'one.txt', '--to', 'store'])

    # reconfigure w/ local overwrite:
    repo.config.add("remote.store.ora-push-url", store_url, scope='local')
    # push works now:
    repo.call_annex(['copy', 'one.txt', '--to', 'store'])

    # remove again (config and file from store)
    repo.call_annex(['move', 'one.txt', '--from', 'store'])
    repo.config.unset("remote.store.ora-push-url", scope='local')
    repo.call_annex(['fsck', '-f', 'store'])
    known_sources = repo.whereis('one.txt')
    assert_in(here_uuid, known_sources)
    assert_not_in(store_uuid, known_sources)

    # reconfigure (this time committed)
    init_opts = common_init_opts + ['url={}'.format(store_url),
                                    'push-url={}'.format(store_url)]
    repo.enable_remote('store', options=init_opts)

    # push works now:
    repo.call_annex(['copy', 'one.txt', '--to', 'store'])
    known_sources = repo.whereis('one.txt')
    assert_in(here_uuid, known_sources)
    assert_in(store_uuid, known_sources)


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732, refactored,
# and patched.
# Skipping on adjusted branch as a proxy for crippledFS. Write permissions of
# the owner on a directory can't be revoked on VFAT. "adjusted branch" is a
# bit broad but covers the CI cases. And everything RIA/ORA doesn't currently
# properly run on crippled/windows anyway. Needs to be more precise when
# RF'ing will hopefully lead to support on windows in principle.
@skip_if_adjusted_branch
def _test_permission(io, store_url, local_ria_store_path, ds):

    # Test whether ORA correctly revokes and obtains write permissions within
    # the annex object tree. That is: Revoke after ORA pushed a key to store
    # in order to allow the object tree to safely be used with an ephemeral
    # clone. And on removal obtain write permissions, like annex would
    # internally on a drop (but be sure to restore if something went wrong).

    populate_dataset(ds)
    ds.save()
    assert_repo_status(ds.path)
    testfile = 'one.txt'

    store_url_path = PurePosixPath(urlparse(store_url).path)

    create_store(io, store_url_path, '1')
    create_ds_in_store(io, store_url_path, ds.id, '2', '1')

    _, _, obj_tree = get_layout_locations(1, local_ria_store_path, ds.id)
    assert_true(obj_tree.is_dir())
    file_key_in_store = obj_tree / 'X9' / '6J' / 'MD5E-s8--7e55db001d319a94b0b713529a756623.txt' / 'MD5E-s8--7e55db001d319a94b0b713529a756623.txt'

    init_opts = common_init_opts + [f'url={"ria+" + store_url}']
    ds.repo.init_remote('store', options=init_opts)

    store_uuid = ds.siblings(
        name='store',
        return_type='item-or-list'
    )['annex-uuid']
    here_uuid = ds.siblings(
        name='here',
        return_type='item-or-list'
    )['annex-uuid']

    known_sources = ds.repo.whereis(testfile)
    assert_in(here_uuid, known_sources)
    assert_not_in(store_uuid, known_sources)
    assert_false(file_key_in_store.exists())

    ds.repo.call_annex(['copy', testfile, '--to', 'store'])
    known_sources = ds.repo.whereis(testfile)
    assert_in(here_uuid, known_sources)
    assert_in(store_uuid, known_sources)
    assert_true(file_key_in_store.exists())

    # Revoke write permissions from parent dir in-store to test whether we
    # still can drop (if we can obtain the permissions). Note, that this has
    # no effect on VFAT.
    file_key_in_store.parent.chmod(
        file_key_in_store.parent.stat().st_mode & ~stat.S_IWUSR
    )
    # we can't directly delete; key in store should be protected
    assert_raises(PermissionError, file_key_in_store.unlink)

    # ORA can still drop, since it obtains permission to:
    ds.repo.call_annex(['drop', testfile, '--from', 'store'])
    known_sources = ds.repo.whereis(testfile)
    assert_in(here_uuid, known_sources)
    assert_not_in(store_uuid, known_sources)
    assert_false(file_key_in_store.exists())


def _get_host_from_ssh_server(ssh_server):
    url_parts = urlparse(ssh_server[0])
    return (
        'ssh://'
        + ((url_parts.username + '@') if url_parts.username else '')
        + url_parts.hostname
        + ((':' + str(url_parts.port)) if url_parts.port else '')
        + '/'
    )


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
@slow  # 17sec + ? on travis
@skip_ssh
def test_version_check_ssh(sshserver, existing_dataset):
    store_dir = _random_name('ria-store-')
    store_url = sshserver[0] + '/' + store_dir
    _test_version_check(
        SSHRemoteIO(store_url),
        store_url,
        sshserver[1] / store_dir,
        existing_dataset
    )


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
def test_version_check(tmp_path, existing_dataset):
    _test_version_check(
        LocalIO(),
        'file://' + (tmp_path / 'ria-store').as_uri(),
        tmp_path / 'ria-store',
        existing_dataset
    )


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
@skip_ssh
def test_obtain_permission_ssh(sshserver, existing_dataset):
    store_dir = _random_name('ria-store-')
    store_url = sshserver[0] + '/' + store_dir
    _test_permission(
        SSHRemoteIO(store_url),
        store_url,
        sshserver[1] / store_dir,
        existing_dataset
    )


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
@skip_if_root
def test_obtain_permission_root(tmp_path, existing_dataset):
    _test_permission(
        LocalIO(),
        'file://' + (tmp_path / 'ria-store').as_uri(),
        tmp_path / 'ria-store',
        existing_dataset
    )


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
def test_setup_store_local(tmp_path):
    _test_setup_store(
        LocalIO(),
        'file://' + (tmp_path / 'ria-store').as_uri(),
        tmp_path / 'ria-store',
    )


@skip_ssh
def test_setup_store_ssh(sshserver):
    store_dir = _random_name('ria-store-')
    store_url = sshserver[0] + '/' + store_dir
    _test_setup_store(
        SSHRemoteIO(store_url),
        store_url,
        sshserver[1] / store_dir,
    )


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
def test_setup_ds_in_store_local(tmp_path):
    _test_setup_ds_in_store(
        LocalIO(),
        'file://' + (tmp_path / 'ria-store').as_uri(),
        tmp_path / 'ria-store',
    )


@skip_ssh
def test_setup_ds_in_store_ssh(sshserver):
    store_dir = _random_name('ria-store-')
    store_url = sshserver[0] + '/' + store_dir
    _test_setup_ds_in_store(
        SSHRemoteIO(store_url),
        store_url,
        sshserver[1] / store_dir,
    )


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
@slow  # 12sec + ? on travis
@skip_ssh
def test_remote_layout_ssh(sshserver, existing_dataset):
    store_dir = _random_name('ria-store-')
    store_url = sshserver[0] + '/' + store_dir
    archive_store_dir = _random_name('ria-archive-store-')
    archive_store_url = sshserver[0] + '/' + archive_store_dir
    _test_remote_layout(
        SSHRemoteIO(store_url),
        store_url,
        archive_store_url,
        sshserver[1] / store_dir,
        sshserver[1] / archive_store_dir,
        existing_dataset,
    )


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
def test_remote_layout(tmp_path, existing_dataset):
    _test_remote_layout(
        LocalIO(),
        'file://' + (tmp_path / 'ria-store').as_uri(),
        'file://' + (tmp_path / 'ria-archive-store').as_uri(),
        tmp_path / 'ria-store',
        tmp_path / 'ria-archive-store',
        existing_dataset,
    )


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
# PATCH remove @know_failure_windows
@skip_ssh
def test_initremote_basic_sshurl(sshserver, tmp_path, existing_dataset):
    store_dir = _random_name('ria-store-')
    store_url = sshserver[0] + '/' + store_dir
    _test_initremote_basic(
        SSHRemoteIO(store_url),
        store_url,
        sshserver[1] / store_dir,
        existing_dataset,
        tmp_path / _random_name('link-'),
    )


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
# PATCH remove @know_failure_windows
def test_initremote_basic_fileurl(tmp_path, existing_dataset):
    store_dir = _random_name('ria-store-')
    _test_initremote_basic(
        LocalIO(),
        'file://' + (tmp_path / store_dir).as_uri(),
        tmp_path / store_dir,
        existing_dataset,
        tmp_path / 'link',
    )


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
# https://github.com/datalad/datalad/issues/6160
@known_failure_windows
def test_initremote_basic_httpurl(http_server, tmp_path, existing_dataset):
    # TODO: add a test for https
    _test_initremote_basic(
        LocalIO(),
        http_server.url,
        http_server.path,
        existing_dataset,
        tmp_path / _random_name('link-'),
        _local_path2pure_posix_path(http_server.path),
    )


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
@turtle
# PATCH remove @know_failure_windows
@skip_ssh
def test_gitannex_ssh(sshserver):
    _test_gitannex(_get_host_from_ssh_server(sshserver))


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
@slow  # 41sec on travis
def test_gitannex_local():
    _test_gitannex(None)


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
# TODO?: make parametric again on _test_ria_postclonecfg
# PATCH remove @know_failure_windows
@slow  # 14 sec on travis
@known_failure_windows  # https://github.com/datalad/datalad/issues/5134
def test_ria_postclonecfg(sshserver):

    if not has_symlink_capability():
        # This is needed to create an ORA remote using an URL for upload,
        # that is then invalidated later on (delete the symlink it's based on).
        raise SkipTest("Can't create symlinks")

    lcl_dir = _random_name('lcl-')
    store_dir = _random_name('ria-store-')
    store2_dir = _random_name('ria-store-2-')

    lcl = sshserver[1] / lcl_dir
    store = sshserver[1] / store_dir
    store2 = sshserver[1] / store2_dir

    lcl.mkdir(parents=True)
    store.mkdir(parents=True)
    store2.mkdir(parents=True)

    #lcl_url = sshserver[0] + '/' + lcl_dir
    store_url = sshserver[0] + '/' + store_dir
    #store2_url = sshserver[0] + '/' + store2_dir

    id = _postclonetest_prepare(lcl, store, store2)

    # test cloning via ria+file://
    _test_ria_postclonecfg(
        get_local_file_url(store, compatibility='git'), id
    )

    # Note: HTTP disabled for now. Requires proper implementation in ORA
    #       remote. See
    # https://github.com/datalad/datalad/pull/4203#discussion_r410284649

    # # test cloning via ria+http://
    # with HTTPPath(store) as url:
    #     yield _test_ria_postclonecfg, url, id

    # test cloning via ria+ssh://
    skip_ssh(_test_ria_postclonecfg)(store_url, id)


# taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
@with_tempfile
@with_tempfile
def _test_ria_postclonecfg(url, dsid, clone_path, superds):
    # Test cloning from RIA store while ORA special remote autoenabling failed
    # due to an invalid URL from the POV of the cloner.
    # Origin's git-config-file should contain the UUID to enable. This needs to
    # work via HTTP, SSH and local cloning.

    # Autoenabling should fail initially by git-annex-init and we would report
    # on INFO level. Only postclone routine would deal with it.
    with swallow_logs(
            new_level=logging.INFO,
            name='datalad.core.distributed.clone',
    ) as cml:
        # First, the super ds:
        riaclone = clone('ria+{}#{}'.format(url, dsid), clone_path)
        cml.assert_logged(msg="access to 1 dataset sibling store-storage not "
                              "auto-enabled",
                          level="INFO",
                          regex=False)

    # However, we now can retrieve content since clone should have enabled the
    # special remote with new URL (or origin in case of HTTP).
    res = riaclone.get('test.txt')
    assert_result_count(res, 1,
                        status='ok',
                        path=str(riaclone.pathobj / 'test.txt'),
                        message="from {}...".format(DEFAULT_REMOTE
                                                    if url.startswith('http')
                                                    else "store-storage"))

    # Second ORA remote is enabled and not reconfigured:
    untouched_remote = riaclone.siblings(name='anotherstore-storage',
                                         return_type='item-or-list')
    assert_not_is_instance(untouched_remote, list)
    untouched_url = riaclone.repo.get_special_remotes()[
        untouched_remote['annex-uuid']]['url']
    ok_(untouched_url.startswith("ria+file://"))
    ok_(not untouched_url.startswith("ria+{}".format(url)))

    # publication dependency was set for store-storage but not for
    # anotherstore-storage:
    eq_(riaclone.config.get(f"remote.{DEFAULT_REMOTE}.datalad-publish-depends",
                            get_all=True),
        "store-storage")

    # same thing for the sub ds (we don't need a store-url and id - get should
    # figure those itself):
    with swallow_logs(
            new_level=logging.INFO,
            name='datalad.core.distributed.clone',
    ) as cml:
        riaclonesub = riaclone.get(
            op.join('subdir', 'subds'), get_data=False,
            result_xfm='datasets', return_type='item-or-list')
        cml.assert_logged(msg="access to 1 dataset sibling store-storage not "
                              "auto-enabled",
                          level="INFO",
                          regex=False)
    res = riaclonesub.get('testsub.txt')
    assert_result_count(res, 1,
                        status='ok',
                        path=str(riaclonesub.pathobj / 'testsub.txt'),
                        message="from {}...".format(DEFAULT_REMOTE
                                                    if url.startswith('http')
                                                    else "store-storage"))

    # publication dependency was set for store-storage but not for
    # anotherstore-storage:
    eq_(riaclonesub.config.get(f"remote.{DEFAULT_REMOTE}.datalad-publish-depends",
                               get_all=True),
        "store-storage")

    # finally get the plain git subdataset.
    # Clone should figure to also clone it from a ria+ URL
    # (subdataset-source-candidate), notice that there wasn't an autoenabled ORA
    # remote, but shouldn't stumble upon it, since it's a plain git.
    res = riaclone.get(op.join('subdir', 'subgit', 'testgit.txt'))
    assert_result_count(res, 1, status='ok', type='dataset', action='install')
    assert_result_count(res, 1, status='notneeded', type='file')
    assert_result_count(res, 2)
    # no ORA remote, no publication dependency:
    riaclonesubgit = Dataset(riaclone.pathobj / 'subdir' / 'subgit')
    eq_(riaclonesubgit.config.get(f"remote.{DEFAULT_REMOTE}.datalad-publish-depends",
                                  get_all=True),
        None)

    # Now, test that if cloning into a dataset, ria-URL is preserved and
    # post-clone configuration is triggered again, when we remove the subds and
    # retrieve it again via `get`:
    ds = Dataset(superds).create()
    ria_url = 'ria+{}#{}'.format(url, dsid)
    ds.clone(ria_url, 'sub')
    sds = ds.subdatasets('sub')
    eq_(len(sds), 1)
    eq_(sds[0]['gitmodule_datalad-url'], ria_url)
    assert_repo_status(ds.path)
    ds.drop('sub', what='all', reckless='kill', recursive=True)
    assert_repo_status(ds.path)

    # .gitmodules still there:
    sds = ds.subdatasets('sub')
    eq_(len(sds), 1)
    eq_(sds[0]['gitmodule_datalad-url'], ria_url)
    # get it again:

    # Autoenabling should fail initially by git-annex-init and we would report
    # on INFO level. Only postclone routine would deal with it.
    with swallow_logs(
            new_level=logging.INFO,
            name='datalad.core.distributed.clone',
    ) as cml:
        ds.get('sub', get_data=False)
        cml.assert_logged(msg="access to 1 dataset sibling store-storage not "
                              "auto-enabled",
                          level="INFO",
                          regex=False)

    subds = Dataset(ds.pathobj / 'sub')
    # special remote is fine:
    res = subds.get('test.txt')
    assert_result_count(res, 1,
                        status='ok',
                        path=str(subds.pathobj / 'test.txt'),
                        message="from {}...".format(DEFAULT_REMOTE
                                                    if url.startswith('http')
                                                    else "store-storage"))



"""Constraints for particular formats or protocols"""

# allow for |-type UnionType declarations
from __future__ import annotations

from json import loads
import re
from urllib.parse import (
    urlparse,
    ParseResult,
)

from .base import Constraint


class EnsureJSON(Constraint):
    """Ensures that string is JSON formatted and can be deserialized.
    """
    def __init__(self):
        super().__init__()

    def __call__(self, value: str):
        try:
            return loads(value)
        except Exception as e:
            self.raise_for(
                value,
                str(e),
            )

    def short_description(self):
        return 'JSON'


class EnsureURL(Constraint):
    """Ensures that a string is a valid URL with a select set of components

    and/or:

    - does not contain certain components
    - matches a particular regular expression

    Given that a large variety of strings are also a valid URL, a typical use
    of this constraint would involve using a `required=['scheme']` setting.

    All URL attribute names supported by `urllib.parse.urlparse()` are also
    supported here: scheme, netloc, path, params, query, fragment, username,
    password, hostname, port.

    .. seealso::
      https://docs.python.org/3/library/urllib.parse.html#urllib.parse.urlparse
    """
    def __init__(
        self,
        required: list | None = None,
        forbidden: list | None = None,
        match: str | None = None,
    ):
        """
        Parameters
        ----------
        required: list, optional
          List of any URL component names as recognized by ``urlparse()``,
          such as ``scheme``, ``netloc``, ``path``, ``params``, ``query``,
          ``fragment``, ``username``, ``password``, ``hostname``, ``port``
        forbidden: list, optional
          Like ``required`` but names URL components that must not be present
        match: str, optional
          Regular expression that the URL must match
        """
        self._required = required
        self._forbidden = forbidden
        self._match_exp = re.compile(match) if match else None
        super().__init__()

    def __call__(self, value: str) -> str:
        self._validate_parsed(value)
        # return the str here, see EnsureParsedURL for an alternative
        return value

    def _validate_parsed(self, value: str) -> ParseResult:
        if not isinstance(value, str):
            self.raise_for(value, 'not a string')
        if self._match_exp and not self._match_exp.match(value):
            self.raise_for(
                value,
                'does not match expression {match_expression!r}',
                match_expression=self._match_exp.pattern,
            )
        parsed = urlparse(value, scheme='', allow_fragments=True)
        for r in (self._required or []):
            if not getattr(parsed, r, None):
                self.raise_for(
                    value,
                    'URL is missing {component!r} component',
                    component=r,
                )
        for f in (self._forbidden or []):
            if getattr(parsed, f, None):
                self.raise_for(
                    value,
                    'URL has forbidden {component!r} component',
                    component=f,
                )
        return parsed

    def short_description(self):
        return 'URL{}{}{}{}'.format(
            f' with required {self._required}' if self._required else '',
            ' and' if self._required and self._forbidden else '',
            f' with no {self._forbidden}' if self._forbidden else '',
            ' component(s)' if self._required or self._forbidden else '',
        )


class EnsureParsedURL(EnsureURL):
    """Like `EnsureURL`, but returns a parsed URL"""
    def __call__(self, value: str) -> ParseResult:
        return self._validate_parsed(value)



import logging
import io

from datalad_next.tests import (
    assert_raises,
    eq_,
)
from datalad_next.utils import swallow_outputs

from ..base import (
    BackendError,
    Master,
    NotLinkedError,
    Protocol,
    ProtocolError,
    UnsupportedRequest,
)


class FakeBackend(object):
    def error(self, message):
        raise ValueError(message)

    def gen_key(self, val):
        raise BackendError('not worky-worky')

    def verify_content(self, f, key):
        raise BackendError('not worky-worky')

    def can_verify(self):
        raise RuntimeError('intentional blow')


def test_protocol():
    """Test essential protocol (error) behavior"""
    p = Protocol(FakeBackend())
    # no empty lines
    assert_raises(ProtocolError, p.command, '')
    # version check
    eq_(p.command('GETVERSION'), 'VERSION 1')
    # unknown command
    assert_raises(UnsupportedRequest, p.command, 'GOTNOCLUE')
    # pass arg to command method
    with assert_raises(ValueError) as ve:
        p.command('ERROR my message')
        eq_(str(ve.exception), 'my message')
    # backend failure is communicated for key generation
    eq_(p.command('GENKEY for-some-file'), 'GENKEY-FAILURE not worky-worky')
    # and for key verification
    eq_(p.command('VERIFYKEYCONTENT for-some-file mykey'), 'VERIFYKEYCONTENT-FAILURE')


def test_master():
    master = Master()
    assert_raises(NotLinkedError, master.Listen)
    master.LinkBackend(FakeBackend())
    with swallow_outputs() as cmo:
        master.Listen(io.StringIO('GETVERSION'))
        eq_(cmo.out, 'VERSION 1\n')
    with swallow_outputs() as cmo:
        master.Listen(io.StringIO('FUNKY'))
        # comes with a DEBUG message showing exactly what went wrong
        eq_(cmo.out, "DEBUG Unknown request 'FUNKY'\nUNSUPPORTED-REQUEST\n")
    assert_raises(SystemExit, master.Listen, io.StringIO('CANVERIFY'))
    with swallow_outputs() as cmo:
        master.progress(15)
        eq_(cmo.out, 'PROGRESS 15\n')





from pathlib import PurePosixPath
import pytest

from datalad.api import download

from datalad_next.tests import skipif_no_network

from ..tarfile import (
    TarfileItem,
    FileSystemItemType,
    iter_tar,
)
from ..utils import compute_multihash_from_fp


@pytest.fixture(scope="session")
def sample_tar_xz(tmp_path_factory):
    """Provides a path to a tarball with file, directory, hard link,
    and soft link. Any file content is '123\n'. The associated hashes
    are:

    md5: ba1f2511fc30423bdbb183fe33f3dd0f
    sha1: a8fdc205a9f19cc1c7507a60c4f01b13d11d7fd0

    Layout::

        ❯ datalad tree --include-files test-archive
        test-archive
        ├── 123.txt -> subdir/onetwothree_again.txt
        ├── 123_hard.txt
        ├── onetwothree.txt
        └── subdir/
            └── onetwothree_again.txt
    """
    path = tmp_path_factory.mktemp("tarfile")
    tfpath = path / 'sample.tar.xz'
    download(
        {'https://github.com/datalad/datalad-next/releases/download/0.1.0/test_archive.tar.xz':
         tfpath},
        result_renderer='disabled',
    )

    yield tfpath

    tfpath.unlink()


@skipif_no_network
def test_iter_tar(sample_tar_xz):
    target_hash = {'SHA1': 'a8fdc205a9f19cc1c7507a60c4f01b13d11d7fd0',
                   'md5': 'ba1f2511fc30423bdbb183fe33f3dd0f'}
    targets = [
        TarfileItem(
            name='test-archive',
            type=FileSystemItemType.directory,
            size=0,
            mtime=1683657433,
            mode=509,
            uid=1000,
            gid=1000),
        TarfileItem(
            name='test-archive/123.txt',
            type=FileSystemItemType.symlink,
            size=0,
            mtime=1683657414,
            mode=511,
            uid=1000,
            gid=1000,
            link_target='subdir/onetwothree_again.txt'),
        TarfileItem(
            name='test-archive/123_hard.txt',
            type=FileSystemItemType.file,
            size=4,
            mtime=1683657364,
            mode=436,
            uid=1000,
            gid=1000,
            link_target=None),
        TarfileItem(
            name='test-archive/subdir',
            type=FileSystemItemType.directory,
            size=0,
            mtime=1683657400,
            mode=509,
            uid=1000,
            gid=1000),
        TarfileItem(
            name='test-archive/subdir/onetwothree_again.txt',
            type=FileSystemItemType.file,
            size=4,
            mtime=1683657400,
            mode=436,
            uid=1000,
            gid=1000,
            link_target=None),
        TarfileItem(
            name='test-archive/onetwothree.txt',
            type=FileSystemItemType.hardlink,
            size=0,
            mtime=1683657364,
            mode=436,
            uid=1000,
            gid=1000,
            link_target='test-archive/123_hard.txt'),
    ]
    ires = []
    for i in iter_tar(sample_tar_xz, fp=True):
        # check that file pointer is usable
        if i.fp:
            assert compute_multihash_from_fp(
                i.fp, ['md5', 'SHA1']) == target_hash
            # we null the file pointers to ease the comparison
            i.fp = None
        ires.append(i)
    # root + subdir, 2 files, softlink, hardlink
    assert 6 == len(ires)
    for t in targets:
        assert t in ires



from datalad_next.tests import (
    assert_in_results,
    assert_raises,
)
from datalad_next.utils import chpwd

from datalad.api import configuration
from datalad_next.exceptions import IncompleteResultsError

# run all -core tests
from datalad.local.tests.test_configuration import *


def test_config_get_global(existing_dataset, tmp_path,
                           no_result_rendering):
    """Make sure `get` does not require a dataset to be present"""
    # enter a tempdir to be confident that there is no dataset around
    with chpwd(str(tmp_path)):
        res = configuration('get', 'user.name')
        assert_in_results(
            res,
            name='user.name',
            status='ok',
        )
    # verify that the dataset method was replaced too
    assert "'get' action can be constrained" \
        in existing_dataset.configuration.__doc__


def test_getset_None(tmp_path, no_result_rendering):
    # enter a tempdir to be confident that there is no dataset around
    with chpwd(str(tmp_path)):
        # set an empty string, this is not the same as `None`
        configuration('set', 'some.item=', scope='global')
        assert_in_results(
            configuration('get', 'some.item'),
            value='',
        )
        # an unset config item is equivalent to `None`
        configuration('unset', 'some.item', scope='global'),
        # retrieving an unset item triggers an exception ...
        assert_raises(
            IncompleteResultsError,
            configuration, 'get', 'some.item')
        # ... because the status of the respective result is "impossible"
        assert_in_results(
            configuration('get', 'some.item',
                          on_failure='ignore'),
            value=None,
            status='impossible',
        )



from pathlib import Path

from datalad_next.runners import call_git_success


def has_initialized_annex(
    path: Path,
) -> bool:
    """Return whether there is an initialized annex for ``path``

    The given ``path`` can be any directory, inside or outside a Git
    repository. ``True`` is returned when the path is found to be
    within a (locally) initialized git-annex repository.

    When this test returns ``True`` it can be expected that no subsequent
    call to an annex command fails with

    `git-annex: First run: git-annex init`

    for this ``path``.
    """
    # this test is about 3ms in MIH's test system.
    # datalad-core tests for a git repo and then for .git/annex, this
    # achieves both in one step (although the test in datalad-core is
    # likely still faster, because it only inspects the filesystem
    # for a few key members of a Git repo. In order for that test to
    # work, though, it has to traverse the filesystem to find a repo root
    # -- if there even is any).
    # also ee https://git-annex.branchable.com/forum/Cheapest_test_for_an_initialized_annex__63__/
    return call_git_success(
        ['config', '--local', 'annex.uuid'],
        cwd=path,
        capture_output=True,
    )



"""git-annex key representation"""

from __future__ import annotations

from dataclasses import dataclass
import re


# BACKEND[-sNNNN][-mNNNN][-SNNNN-CNNNN]--NAME
_annexkey_regex = re.compile(
    '(?P<backend>[A-Z0-9]+)'
    '(|-s(?P<size>[0-9]+))'
    '(|-m(?P<mtime>[0-9]+))'
    '(|-S(?P<chunksize>[0-9]+)-C(?P<chunknumber>[0-9]+))'
    '--(?P<name>.*)$'
)


@dataclass(frozen=True)
class AnnexKey:
    """Representation of a git-annex key

    https://git-annex.branchable.com/internals/key_format/
    """
    name: str
    backend: str
    size: int | None = None
    mtime: int | None = None
    chunksize: int | None = None
    chunknumber: int | None = None

    @classmethod
    def from_str(cls, key: str):
        """Return an ``AnnexKey`` instance from a key string"""
        key_matched = _annexkey_regex.match(key)
        if not key_matched:
            # without a sensible key there is no hope
            raise ValueError(f'{key!r} is not a valid git-annex key')
        return cls(**key_matched.groupdict())

    def __str__(self) -> str:
        return '{backend}{size}{mtime}{chunk}--{name}'.format(
            name=self.name,
            backend=self.backend,
            size=f'-s{self.size}' if self.size else '',
            mtime=f'-m{self.mtime}' if self.mtime else '',
            # if me reading of the spec is correct, the two chunk props
            # can only occur together
            chunk=f'-S{self.chunksize}-C{self.chunknumber}'
            if self.chunknumber else '',
        )



from pathlib import (
    PurePosixPath,
)

import pytest

from datalad_next.utils import rmtree

from ..gittree import (
    GitTreeItem,
    GitTreeItemType,
    iter_gittree,
)


def test_iter_gittree(existing_dataset, no_result_rendering):
    ds = existing_dataset
    is_crippled_fs = ds.repo.is_crippled_fs()

    tracked_items = list(iter_gittree(ds.pathobj, 'HEAD'))
    # without untracked's and no link resolution this is plain and fast
    assert all(
        isinstance(i, GitTreeItem) and i.gitsha and i.gittype
        for i in tracked_items
    )
    # we add a new file and test its expected properties
    probe_name = 'probe.txt'
    # on crippled FS we are testing the managed branch which contains
    # pointer files, not symlinks
    expected_probe_sha = '969921a905b411137fce77c104954f742e1ee6c7' \
        if is_crippled_fs \
        else '7c38bf0378c31f8696e5869e7828a32c9dc2684e'
    probe = ds.pathobj / 'subdir' / probe_name
    probe.parent.mkdir()
    probe.write_text('probe')
    ds.save()
    assert any(
        # let's query a Path instance here, to get that covered too
        i.path == PurePosixPath(f'subdir/{probe_name}')
        and i.gitsha == expected_probe_sha
        and (
            i.gittype in (GitTreeItemType.file, GitTreeItemType.executablefile)
            if is_crippled_fs
            else i.gittype == GitTreeItemType.symlink
        )
        for i in iter_gittree(ds.pathobj, 'HEAD')
    )
    if not is_crippled_fs:
        # if we check the prior version, we do not see it (hence the
        # tree-ish passing is working
        assert not any(
            i.path == PurePosixPath(f'subdir/{probe_name}')
            for i in iter_gittree(ds.pathobj, 'HEAD~1')
        )

    # if we disable recursion, the probe is not listed, but its
    # parent dir is
    tracked_toplevel_items = list(
        iter_gittree(ds.pathobj, 'HEAD', recursive='no'))
    assert not any(
        i.name == f'subdir/{probe_name}'
        for i in tracked_toplevel_items
    )
    assert any(
        i.name == 'subdir'
        and (True if is_crippled_fs
             else 'eb4aa65f42b90178837350571a227445b996cf90')
        and i.gittype == GitTreeItemType.directory
        for i in tracked_toplevel_items
    )
    # iterating on a subdir does constrain the report
    tracked_subdir_items = list(iter_gittree(probe.parent, 'HEAD'))
    assert len(tracked_subdir_items) == 1
    probe_item = tracked_subdir_items[0]
    assert probe_item.name == probe_name
    assert probe_item.gitsha == expected_probe_sha


def test_name_starting_with_tab(existing_dataset, no_result_rendering):
    ds = existing_dataset
    if ds.repo.is_crippled_fs():
        pytest.skip("not applicable on crippled filesystems")
    tabbed_file_name = "\ttab.txt"
    tabbed_name = ds.pathobj / tabbed_file_name
    tabbed_name.write_text('name of this file starts with a tab')
    ds.save()
    iter_names = [item.path for item in iter_gittree(ds.pathobj, 'HEAD')]
    assert PurePosixPath(tabbed_file_name) in iter_names


def test_iter_gittree_empty(existing_dataset, no_result_rendering):
    ds = existing_dataset
    rmtree(ds.pathobj / '.datalad')
    (ds.pathobj / '.gitattributes').unlink()
    ds.save()
    assert len(ds.status()) == 0
    all_items = list(iter_gittree(ds.pathobj, 'HEAD'))
    assert len(all_items) == 0



"""Configuration query and manipulation

This modules provides the central ``ConfigManager`` class.

.. currentmodule:: datalad_next.config
.. autosummary::
   :toctree: generated

   ConfigManager
"""

from datalad.config import ConfigManager



from datalad.tests.test_config import *


# this datalad-core test is causing a persistent git config modification
# this is not legal on datalad-next, we must wrap and protect
_test_cross_cfgman_update = test_cross_cfgman_update


def test_cross_cfgman_update(datalad_cfg, tmp_path):
    _test_cross_cfgman_update(tmp_path)



"""Base classes for constraints and their logical connectives
"""

from __future__ import annotations

__docformat__ = 'restructuredtext'

__all__ = ['Constraint', 'AllOf', 'AnyOf', 'DatasetParameter']

from .exceptions import ConstraintError


class DatasetParameter:
    """Utility class to report an original and resolve dataset parameter value

    This is used by `EnsureDataset` to be able to report the original argument
    semantics of a dataset parameter to a receiving command. It is consumed
    by any ``Constraint.for_dataset()``.

    The original argument is provided via the `original` property.
    A corresponding `Dataset` instance is provided via the `ds` property.
    """
    def __init__(self, original, ds):
        self.original = original
        self.ds = ds

    def __repr__(self):
        return f'{self.__class__.__name__}({self.original}, {self.ds})'


class Constraint:
    """Base class for value coercion/validation.

    These classes are also meant to be able to generate appropriate
    documentation on an appropriate parameter value.
    """
    def __str__(self):
        """Rudimentary self-description"""
        return f"constraint: {self.short_description()}"

    def __repr__(self):
        """Rudimentary repr to avoid default scary to the user Python repr"""
        return f"{self.__class__.__name__}()"

    def raise_for(self, value, msg, **ctx) -> None:
        """Convenience method for raising a ``ConstraintError``

        The parameters are identical to those of ``ConstraintError``. This
        method merely passes the ``Constraint`` instance as ``self`` to the
        constructor.
        """
        if ctx:
            raise ConstraintError(self, value, msg, ctx)
        else:
            raise ConstraintError(self, value, msg)

    def __and__(self, other):
        return AllOf(self, other)

    def __or__(self, other):
        return AnyOf(self, other)

    def __call__(self, value):
        # do any necessary checks or conversions, potentially catch exceptions
        # and generate a meaningful error message
        raise NotImplementedError("abstract class")

    @property
    def input_synopsis(self) -> str:
        """Returns brief, single line summary of valid input for a constraint

        This information is user-facing, and to be used in any place where
        space is limited (tooltips, usage summaries, etc).

        If possible, the synopsis should be written in a UI/API-agnostic
        fashion. However, if this is impossible or leads to imprecisions or
        confusion, it should focus on use within Python code and with Python
        data types. Tailored documentation can be provided via the
        ``WithDescription`` wrapper.
        """
        # return the legacy short description for now
        return self.short_description()

    @property
    def input_description(self) -> str:
        """Returns full description of valid input for a constraint

        Like ``input_synopsis`` this information is user-facing. In contrast,
        to the synopsis there is length/line limit. Nevertheless, the
        information should be presented in a compact fashion that avoids
        needless verbosity. If possible, a single paragraph is a good format.
        If multiple paragraphs are necessary, they should be separated by
        a single, empty line.

        Rendering code may indent, or rewrap the text, so no line-by-line
        formatting will be preserved.

        If possible, the synopsis should be written in a UI/API-agnostic
        fashion. However, if this is impossible or leads to imprecisions or
        confusion, it should focus on use within Python code and with Python
        data types. Tailored documentation can be provided via the
        ``WithDescription`` wrapper.
        """
        # return the legacy short description for now
        return self.long_description()

    def long_description(self):
        """This method is deprecated. Use ``input_description`` instead"""
        # return meaningful docs or None
        # used as a comprehensive description in the parameter list
        return self.short_description()

    def short_description(self):
        """This method is deprecated. Use ``input_synopsis`` instead"""
        # return meaningful docs or None
        # used as a condensed primer for the parameter lists
        raise NotImplementedError("abstract class")

    def for_dataset(self, dataset: DatasetParameter) -> Constraint:
        """Return a constraint-variant for a specific dataset context

        The default implementation returns the unmodified, identical
        constraint. However, subclasses can implement different behaviors.
        """
        return self


class _MultiConstraint(Constraint):
    """Helper class to override the description methods to reported
    multiple constraints
    """
    def __init__(self, *constraints):
        # TODO Why is EnsureNone needed? Remove if possible
        from .basic import EnsureNone
        self._constraints = [
            EnsureNone() if c is None else c for c in constraints
        ]

    def __repr__(self):
        creprs = ', '.join(f'{c!r}' for c in self.constraints)
        return f"{self.__class__.__name__}({creprs})"

    @property
    def constraints(self):
        return self._constraints

    def _get_description(self, attr: str, operation: str) -> str:
        cs = [
            getattr(c, attr)()
            for c in self.constraints
            if hasattr(c, attr)
        ]
        cs = [c for c in cs if c is not None]
        doc = f' {operation} '.join(cs)
        if len(cs) > 1:
            return f'({doc})'
        else:
            return doc


class AnyOf(_MultiConstraint):
    """Logical OR for constraints.

    An arbitrary number of constraints can be given. They are evaluated in the
    order in which they were specified. The value returned by the first
    constraint that does not raise an exception is the global return value.

    Documentation is aggregated for all alternative constraints.
    """
    def __init__(self, *constraints):
        """
        Parameters
        ----------
        *constraints
           Alternative constraints
        """
        super().__init__(*constraints)

    def __or__(self, other):
        constraints = list(self.constraints)
        if isinstance(other, AnyOf):
            constraints.extend(other.constraints)
        else:
            constraints.append(other)
        return AnyOf(*constraints)

    def __call__(self, value):
        e_list = []
        for c in self.constraints:
            try:
                return c(value)
            except Exception as e:
                e_list.append(e)
        self.raise_for(
            value,
            # plural OK, no sense in having 1 "alternative"
            'does not match any of {n_alternatives} alternatives\n'
            '{__itemized_causes__}',
            # if any exception would be a ConstraintError
            # this would not be needed, because they
            # know the underlying constraint
            constraints=self.constraints,
            n_alternatives=len(self.constraints),
            __caused_by__=e_list,
        )

    def long_description(self):
        return self._get_description('long_description', 'or')

    def short_description(self):
        return self._get_description('short_description', 'or')


class AllOf(_MultiConstraint):
    """Logical AND for constraints.

    An arbitrary number of constraints can be given. They are evaluated in the
    order in which they were specified. The return value of each constraint is
    passed an input into the next. The return value of the last constraint
    is the global return value. No intermediate exceptions are caught.

    Documentation is aggregated for all constraints.
    """
    def __init__(self, *constraints):
        """
        Parameters
        ----------
        *constraints
           Constraints all of which must be satisfied
        """
        super().__init__(*constraints)

    def __and__(self, other):
        constraints = list(self.constraints)
        if isinstance(other, AllOf):
            constraints.extend(other.constraints)
        else:
            constraints.append(other)
        return AllOf(*constraints)

    def __call__(self, value):
        for c in (self.constraints):
            value = c(value)
        return value

    def long_description(self):
        return self._get_description('long_description', 'and')

    def short_description(self):
        return self._get_description('short_description', 'and')


# keep for backward compatibility
Constraints = AllOf
AltConstraints = AnyOf




"""Report on the content of ZIP file

The main functionality is provided by the :func:`iter_zip()` function.
"""

from __future__ import annotations

import datetime
from functools import cached_property
import time
import zipfile
from dataclasses import dataclass
from pathlib import (
    Path,
    PurePosixPath,
)
from typing import Generator

from .utils import (
    FileSystemItem,
    FileSystemItemType,
)


@dataclass
class ZipfileItem(FileSystemItem):
    name: str

    @cached_property
    def path(self) -> PurePosixPath:
        """Returns the item name as a ``PurePosixPath`` instance

        ZIP uses POSIX paths as item identifiers from version 6.3.3 onwards.
        Not all POSIX paths are legal paths on non-POSIX file systems or
        platforms.  Therefore we cannot use a platform-dependent
        ``PurePath``-instance to address ZIP-file items, anq we use
        ``PurePosixPath``-instances instead.
        """
        return PurePosixPath(self.name)


def iter_zip(
        path: Path,
        *,
        fp: bool = False,
) -> Generator[ZipfileItem, None, None]:
    """Uses the standard library ``zipfile`` module to report on ZIP-files

    A ZIP archive can represent more or less the full bandwidth of file system
    properties, therefore reporting on archive members is implemented
    similar to :func:`~datalad_next.iter_collections.directory.iter_dir()`.
    The iterator produces an :class:`ZipfileItem` instance with standard
    information on file system elements, such as ``size``, or ``mtime``.

    Parameters
    ----------
    path: Path
      Path of the ZIP archive to report content for (iterate over).
    fp: bool, optional
      If ``True``, each file-type item includes a file-like object
      to access the file's content. This file handle will be closed
      automatically when the next item is yielded or the function
      returns.

    Yields
    ------
    :class:`ZipfileItem`
      The ``name`` attribute of an item is a ``str`` with the corresponding
      archive member name (in POSIX conventions).
    """
    with zipfile.ZipFile(path, mode='r') as zip_file:
        for zip_info in zip_file.infolist():
            item = _get_zipfile_item(zip_info)
            if fp and item.type == FileSystemItemType.file:
                with zip_file.open(zip_info) as amfp:
                    item.fp = amfp
                    yield item
            else:
                yield item


def _get_zipfile_item(zip_info: zipfile.ZipInfo) -> ZipfileItem:
    return ZipfileItem(
        name=zip_info.filename,
        type=FileSystemItemType.directory if zip_info.is_dir()
        else FileSystemItemType.file,
        size=zip_info.file_size,
        mtime=time.mktime(
            datetime.datetime(*zip_info.date_time).timetuple()
        )
    )



"""ZIP archive operation handler"""

from __future__ import annotations

import logging
import zipfile
from contextlib import contextmanager
from pathlib import (
    Path,
    PurePosixPath,
)
from typing import (
    Generator,
    IO,
)
from zipfile import ZipInfo

from datalad_next.config import ConfigManager
# TODO we might just want to do it in reverse:
# move the code of `iter_zip` in here and have it call
# `ZipArchiveOperations(path).__iter__()` instead.
# However, the flexibility to have `iter_zip()` behave
# differently depending on parameters (fp=True/False)
# is nice, and `__iter__()` only has `self`, such that
# any customization would need to be infused in the whole
# class. Potentially cumbersome.
from datalad_next.iter_collections import (
    ZipfileItem,
    iter_zip,
)
from .base import ArchiveOperations


lgr = logging.getLogger('datalad.ext.next.archive_operations.zipfile')


class ZipArchiveOperations(ArchiveOperations):
    """Handler for a ZIP archive on a local file system
    """
    def __init__(self,
                 location: Path,
                 *,
                 cfg: ConfigManager | None = None,
                 **kwargs):
        """
        Parameters
        ----------
        location: Path
          ZIP archive location
        cfg: ConfigManager, optional
          A config manager instance that is consulted for any supported
          configuration items
        **kwargs: dict
          Keyword arguments that are passed to zipfile.ZipFile-constructor
        """
        super().__init__(location, cfg=cfg)

        self.zipfile_kwargs = kwargs
        # Consider supporting file-like for `location`,
        # see zipfile.ZipFile(file_like_object)
        self._zipfile_path = location
        self._zipfile: zipfile.ZipFile | None = None

    @property
    def zipfile(self) -> zipfile.ZipFile:
        """Access to the wrapped ZIP archive as a ``zipfile.ZipFile``"""
        if self._zipfile is None:
            self._zipfile = zipfile.ZipFile(
                self._zipfile_path,
                **self.zipfile_kwargs
            )
        return self._zipfile

    def close(self) -> None:
        """Calls `.close()` on the underlying ``zipfile.ZipFile`` instance"""
        if self._zipfile:
            self._zipfile.close()
            self._zipfile = None

    @contextmanager
    def open(
        self,
        item: str | PurePosixPath | ZipInfo,
        **kwargs,
    ) -> Generator[IO | None, None, None]:
        """Context manager, returning an open file for a member of the archive.

        The file-like object will be closed when the context-handler
        exits.

        This method can be used in conjunction with ``__iter__`` to read any
        file from an archive::

            with ZipArchiveOperations(archive_path) as zf:
                for item in zf:
                    if item.type != FileSystemItemType.file:
                        continue
                    with zf.open(item.name) as fp:
                        ...

        Parameters
        ----------
        item: str | PurePosixPath | zipfile.ZipInfo
          Name, path, or ZipInfo-instance that identifies an item in the
          zipfile
        kwargs: dict
          Keyword arguments that will be used for ZipFile.open()

        Returns
        -------
        IO
          A file-like object to read bytes from the item or to write bytes
          to the item.
        """
        with self.zipfile.open(_anyzipid2membername(item), **kwargs) as fp:
            yield fp

    def __contains__(self, item: str | PurePosixPath | ZipInfo) -> bool:
        try:
            self.zipfile.getinfo(_anyzipid2membername(item))
            return True
        except KeyError:
            return False

    def __iter__(self) -> Generator[ZipfileItem, None, None]:
        # if fp=True is needed, either `iter_zip()` can be used
        # directly, or `ZipArchiveOperations.open`
        yield from iter_zip(self._zipfile_path, fp=False)


def _anyzipid2membername(item: str | PurePosixPath | ZipInfo) -> str:
    """Convert any supported archive member ID for ``zipfile.open|getinfo()``
    """
    if isinstance(item, ZipInfo):
        return item.filename
    elif isinstance(item, PurePosixPath):
        return item.as_posix()
    return item



from __future__ import annotations

from dataclasses import dataclass
from pathlib import (
    Path,
    PurePosixPath,
)
from typing import Generator

import pytest

from datalad_next.iter_collections import FileSystemItemType

from ..zipfile import ZipArchiveOperations


@dataclass
class _TestArchive:
    path: Path
    item_count: int
    content: bytes
    target_hash: dict[str, str]


@pytest.fixture(scope='session')
def structured_sample_zip(sample_zip) -> Generator[_TestArchive, None, None]:
    yield _TestArchive(
        path=sample_zip,
        item_count=4,
        content=b'zip-123\n',
        target_hash={
            'SHA1': 'b5dfcec4d1b6166067226fae102f7fbcf6bd1bd4',
            'md5': 'd700214df5487801e8ee23d31e60382a',
        }
    )


def test_ziparchive_basics(structured_sample_zip: _TestArchive):
    spec = structured_sample_zip
    # this is intentionally a hard-coded POSIX relpath
    member_name = 'test-archive/onetwothree.txt'
    with ZipArchiveOperations(spec.path) as archive_ops:
        with archive_ops.open(member_name) as member:
            assert member.read() == spec.content
        with archive_ops.open(PurePosixPath(member_name)) as member:
            assert member.read() == spec.content


def test_ziparchive_contain(structured_sample_zip: _TestArchive):
    # this is intentionally a hard-coded POSIX relpath
    member_name = 'test-archive/onetwothree.txt'
    with ZipArchiveOperations(structured_sample_zip.path) as archive_ops:
        assert member_name in archive_ops
        assert PurePosixPath(member_name) in archive_ops
        assert 'bogus' not in archive_ops


def test_ziparchive_iterator(structured_sample_zip: _TestArchive):
    spec = structured_sample_zip
    with ZipArchiveOperations(spec.path) as archive_ops:
        items = list(archive_ops)
        assert len(items) == spec.item_count
        for item in items:
            assert item.name in archive_ops


def test_open(structured_sample_zip: _TestArchive):
    spec = structured_sample_zip
    file_pointer = set()
    with ZipArchiveOperations(spec.path) as zf:
        for item in zf:
            if item.type != FileSystemItemType.file:
                continue
            with zf.open(item.name) as fp:
                file_pointer.add(fp)
                assert fp.read(len(spec.content)) == spec.content
        for fp in file_pointer:
            assert fp.closed is True


def test_open_zipinfo(structured_sample_zip: _TestArchive):
    spec = structured_sample_zip
    with ZipArchiveOperations(spec.path) as zf:
        # get zipfile-native ZipInfo items
        for item in zf.zipfile.infolist():
            if item.filename.endswith('/'):
                # crude approach to skippping non-files
                continue
            with zf.open(item) as fp:
                assert fp.read(len(spec.content)) == spec.content


def test_ziparchive_noncontext(structured_sample_zip: _TestArchive):
    spec = structured_sample_zip
    zip = ZipArchiveOperations(spec.path)
    assert zip.zipfile.filename == str(spec.path)
    zip.close()
    assert zip._zipfile is None



import logging
import os

lgr = logging.getLogger('datalad.utils.specialremote')


def get_specialremote_param_dict(params):
    """
    Parameters
    ----------
    params : list

    Returns
    -------
    dict
    """
    return dict(p.split('=', maxsplit=1) for p in params)


def get_specialremote_credential_properties(params):
    """Determine properties of credentials special remote configuration

    The input is a parameterization as it would be given to
    `git annex initremote|enableremote <name> ...`, or as stored in
    `remote.log`. These parameters are inspected and a dictionary
    of credential properties, suitable for `CredentialManager.query()`
    is returned. This inspection may involve network activity, e.g.
    HTTP requests.

    Parameters
    ----------
    params : list or dict
      Either a list of strings of the format 'param=value', or a dictionary
      with parameter names as keys.

    Returns
    -------
    dict or None
      Credential property name-value mapping. This mapping can be passed to
      `CredentialManager.query()`. If no credential properties could be
      inferred, for example, because the special remote type is not recognized
      `None` is returned.
    """
    if isinstance(params, (list, tuple)):
        params = get_specialremote_param_dict(params)

    props = {}
    # no other way to do this specifically for each supported remote type
    remote_type = params.get('type')
    if remote_type == 'webdav':
        from .http_helpers import get_auth_realm
        from datalad_next.url_operations import HttpUrlOperations
        url = params.get('url')
        if not url:
            return
        url, urlprops = HttpUrlOperations().probe_url(url)
        realm = get_auth_realm(url, urlprops.get('auth'))
        if realm:
            props['realm'] = realm
    else:
        return

    return props or None


def update_specialremote_credential(
        srtype, credman, credname, credprops, credtype_hint=None,
        duplicate_hint=None):
    """
    Parameters
    ----------
    srtype: str
    credman: CredentialManager
    credname: str or Name
    credprops: dict
    """
    if not credname:
        # name could still be None, if this was entered
        # create a default name, and check if it has not been used
        credname = '{type}{udelim}{user}{delim}{realm}'.format(
            type=srtype,
            udelim='-' if 'user' in credprops else '',
            user=credprops.get('user', ''),
            delim='-' if 'realm' in credprops else '',
            realm=credprops.get('realm', ''),
        )
        if credman.get(
                name=credname,
                # give to make legacy credentials accessible
                _type_hint=credtype_hint):
            # this is already in use, do not override
            lgr.warning(
                'The entered credential will not be stored, '
                'a credential with the default name %r already exists.%s',
                credname, f' {duplicate_hint}' if duplicate_hint else '')
            return
    # we have used a credential, store it with updated usage info
    try:
        credman.set(credname, _lastused=True, **credprops)
    except Exception as e:
        from datalad_next.exceptions import CapturedException

        # we do not want to crash for any failure to store a
        # credential
        lgr.warn(
            'Exception raised when storing credential %r %r: %s',
            credname, credprops, CapturedException(e),
        )


# mapping for credential properties for specific special remote
# types. this is unpleasantly non-generic, but only a small
# subset of git-annex special remotes require credentials to be
# given via ENV vars, and all of rclone's handle it internally
specialremote_credential_envmap = dict(
    # it makes no sense to pull a short-lived access token from
    # a credential store, it can be given via AWS_SESSION_TOKEN
    # in any case
    glacier=dict(
        user='AWS_ACCESS_KEY_ID',  # nosec
        secret='AWS_SECRET_ACCESS_KEY'),  # nosec
    s3=dict(
        user='AWS_ACCESS_KEY_ID',  # nosec
        secret='AWS_SECRET_ACCESS_KEY'),  # nosec
    webdav=dict(
        user='WEBDAV_USERNAME',  # nosec
        secret='WEBDAV_PASSWORD'),  # nosec
)


def needs_specialremote_credential_envpatch(remote_type):
    """Returns whether the environment needs to be patched with credentials

    Returns
    -------
    bool
      False, if the special remote type is not recognized as one needing
      credentials, or if there are credentials already present.
      True, otherwise.
    """
    if remote_type not in specialremote_credential_envmap:
        lgr.debug('Special remote type %r not supported for credential setup',
                  remote_type)
        return False

    # retrieve deployment mapping
    env_map = specialremote_credential_envmap[remote_type]
    if all(k in os.environ for k in env_map.values()):
        # the ENV is fully set up
        # let's prefer the environment to behave like git-annex
        lgr.debug(
            'Not deploying credentials for special remote type %r, '
            'already present in environment', remote_type)
        return False

    # no counterevidence
    return True


def get_specialremote_credential_envpatch(remote_type, cred):
    """Create an environment path for a particular remote type and credential

    Returns
    -------
    dict or None
      A dict with all required items to patch the environment, or None
      if not enough information is available, or nothing needs to be patched.
    """
    env_map = specialremote_credential_envmap.get(remote_type, {})
    return {
        # take whatever partial setup the ENV has already
        v: cred[k]
        for k, v in env_map.items()
        if v not in os.environ
    } or None



"""Change the default of ``datalad.annex.retry`` to ``1``

This prevents unconditional retries, and thereby improves the legibility
of errors (now only one error instead of three identical errors).

This change does not override user-settings, only the default.
"""

from datalad.support.extensions import has_config

if has_config('datalad.annex.retry'):
    from datalad.interface.common_cfg import definitions
    retrycfg = definitions['datalad.annex.retry']
    retrycfg['default'] = 1



from . import (
    cli_configoverrides,
    commanderror,
    common_cfg,
    annexrepo,
    configuration,
    create_sibling_ghlike,
    interface_base,
    interface_utils,
    push_to_export_remote,
    push_optimize,
    siblings,
    test_keyring,
    customremotes_main,
    create_sibling_gitlab,
    patch_ria_ora,
    run,
    # the following two patches have been taken verbatim from datalad-ria
    ssh_exec,
    sshconnector,
)



"""DataLad NEXT extension"""

__docformat__ = 'restructuredtext'

__all__ = [
    '__version__',
    'command_suite',
    # TODO: REMOVE FOR V2.0
    'lgr',
]

import logging

from datalad_next._version import __version__

# patch datalad-core
from datalad_next.patches import enabled as enabled  # noqa: PLC0414

# TODO: REMOVE FOR V2.0
lgr = logging.getLogger('datalad.next')

# Defines a datalad command suite.
# This variable must be bound as a setuptools entrypoint
# to be found by datalad
command_suite = (
    # description of the command suite, displayed in cmdline help
    'What is next in DataLad',
    [
        # specification of a command, any number of commands can be defined
        (
            # importable module that contains the command implementation
            'datalad_next.commands.credentials',
            # name of the command class implementation in above module
            'Credentials',
        ),
        (
            # importable module that contains the command implementation
            'datalad_next.commands.create_sibling_webdav',
            # name of the command class implementation in above module
            'CreateSiblingWebDAV',
            # we gotta make this explicit, or the build_support code will
            # not pick it up, due to the dashes in the name
            'create-sibling-webdav',
        ),
        (
            # importable module that contains the command implementation
            'datalad_next.commands.tree',
            # name of the command class implementation in above module
            'TreeCommand',
            # command name (differs from lowercase command class name)
            'tree',
        ),
        (
            'datalad_next.commands.download',
            'Download',
            'download',
        ),
        (
            'datalad_next.commands.ls_file_collection',
            'LsFileCollection',
            'ls-file-collection',
        ),
        (
            'datalad_next.commands.status',
            'Status',
            'next-status',
            'next_status',
        ),
    ],
)


# register additional configuration items in datalad-core
from datalad.support.extensions import register_config  # noqa: E402

from datalad_next.constraints import (  # noqa: E402
    EnsureBool,
    EnsureChoice,
)

register_config(
    'datalad.credentials.repeat-secret-entry',
    'Require entering secrets twice for interactive specification?',
    type=EnsureBool(),
    default=True,
    dialog='yesno',
)
register_config(
    'datalad.credentials.hidden-secret-entry',
    'Hide secret in interactive entry?',
    type=EnsureBool(),
    default=True,
    dialog='yesno',
)
register_config(
    'datalad.clone.url-substitute.webdav',
    'webdav(s):// clone URL substitution',
    description='Convenience conversion of custom WebDAV URLs to '
    "git-cloneable 'datalad-annex::'-type URLs. The 'webdav://' "
    "prefix implies a remote sibling in 'filetree' or 'export' mode "
    'See https://docs.datalad.org/design/url_substitution.html for details',
    dialog='question',
    scope='global',
    default=(
        r',^webdav([s]*)://([^?]+)$,datalad-annex::http\1://\2?type=webdav&encryption=none&exporttree=yes&url={noquery}',
    ),
)
register_config(
    'datalad.runtime.parameter-violation',
    'Perform exhaustive command parameter validation, or fail on first error?',
    type=EnsureChoice('raise-early', 'raise-at-end'),
    default='raise-early',
    dialog='question',
)
register_config(
    'datalad.archivist.legacy-mode',
    'Fall back on legacy ``datalad-archives`` special remote implementation?',
    description='If enabled, all `archivist` special remote operations '
    'fall back onto the legacy ``datalad-archives`` special remote '
    'implementation. This mode is only provided for backward-compatibility. '
    'This legacy implementation unconditionally downloads archive files '
    'completely, and keeps an internal cache of the full extracted archive '
    'around. The implied 200% storage cost overhead for obtaining a complete '
    'dataset can be prohibitive for datasets tracking large amount of data '
    '(in archive files).',
    type=EnsureBool(),
    default=False,
    dialog='yesno',
)



"""API base class"""

from __future__ import annotations


import logging
from functools import partial
from more_itertools import side_effect
from pathlib import Path
from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
)

import datalad
from datalad_next.config import ConfigManager
from datalad_next.utils import log_progress
from datalad_next.utils.multihash import (
    MultiHash,
    NoOpHash,
)


lgr = logging.getLogger('datalad.ext.next.url_operations')


class UrlOperations:
    """Abstraction for operations on URLs

    Support for specific URL schemes can be implemented via sub-classes.
    Such classes must comply with the following conditions:

    - Any configuration look-up must be performed with the `self.cfg`
      property, which is guaranteed to be a `ConfigManager` instance.

    - When downloads are to be supported, implement the `download()` method
      and comply with the behavior described in its documentation.

    This class provides a range of helper methods to aid computation of
    hashes and progress reporting.
    """
    def __init__(self, *, cfg: ConfigManager | None = None):
        """
        Parameters
        ----------
        cfg: ConfigManager, optional
          A config manager instance that implementations will consult for
          any configuration items they may support.
        """
        self._cfg = cfg

    @property
    def cfg(self) -> ConfigManager:

        if self._cfg is None:
            self._cfg = datalad.cfg
        return self._cfg

    def stat(self,
             url: str,
             *,
             credential: str | None = None,
             timeout: float | None = None) -> Dict:
        """Gather information on a URL target, without downloading it

        Returns
        -------
        dict
          A mapping of property names to values of the URL target. The
          particular composition of properties depends on the specific
          URL. A standard property is 'content-length', indicating the
          size of a download.

        Raises
        ------
        UrlOperationsRemoteError
          This exception is raised on any access-related error on the remote
          side, with a summary of the underlying issues as its message.
          It may carry a status code (e.g. HTTP status code) as its
          ``status_code`` property.
          Any underlying exception must be linked via the `__cause__`
          property (e.g. `raise UrlOperationsRemoteError(...) from ...`).
        UrlOperationsInteractionError
        UrlOperationsAuthenticationError
        UrlOperationsAuthorizationError
        UrlOperationsResourceUnknown
          Implementations that can distinguish several remote error types
          beyond indication a general ``UrlOperationsRemoteError``:
          ``UrlOperationsInteractionError`` general issues in communicating
          with the remote side; ``UrlOperationsAuthenticationError`` for
          errors related to (failed) authentication at the remote;
          ``UrlOperationsAuthorizationError`` for (lack of) authorizating
          to access a particular resource of perform a particular operation;
          ``UrlOperationsResourceUnknown`` if the target of an operation
          does not exist.
        TimeoutError
          If `timeout` is given and the operation does not complete within the
          number of seconds that a specified by `timeout`.
        """
        raise NotImplementedError

    def download(self,
                 from_url: str,
                 to_path: Path | None,
                 *,
                 credential: str | None = None,
                 hash: list[str] | None = None,
                 timeout: float | None = None) -> Dict:
        """Download from a URL to a local file or stream to stdout

        Parameters
        ----------
        from_url: str
          Valid URL with any scheme supported by a particular implementation.
        to_path: Path or None
          A local platform-native path or `None`. If `None` the downloaded
          data is written to `stdout`, otherwise it is written to a file
          at the given path. The path is assumed to not exist. Any existing
          file will be overwritten.
        credential: str, optional
          The name of a dedicated credential to be used for authentication
          in order to perform the download. Particular implementations may
          or may not require or support authentication. They also may or
          may not support automatic credential lookup.
        hash: list(algorithm_names), optional
          If given, must be a list of hash algorithm names supported by the
          `hashlib` module. A corresponding hash will be computed simultaneously
          to the download (without reading the data twice), and included
          in the return value.
        timeout: float, optional
          If given, specifies a timeout in seconds. If the operation is not
          completed within this time, it will raise a `TimeoutError`-exception.
          If timeout is None, the operation will never timeout.

        Returns
        -------
        dict
          A mapping of property names to values for the completed download.
          If `hash` algorithm names are provided, a corresponding key for
          each algorithm is included in this mapping, with the hexdigest
          of the corresponding checksum as the value.

        Raises
        ------
        UrlOperationsRemoteError
          This exception is raised on any deletion-related error on the remote
          side, with a summary of the underlying issues as its message.
          It may carry a status code (e.g. HTTP status code) as its
          ``status_code`` property.
          Any underlying exception must be linked via the `__cause__`
          property (e.g. `raise UrlOperationsRemoteError(...) from ...`).
        UrlOperationsInteractionError
        UrlOperationsAuthenticationError
        UrlOperationsAuthorizationError
        UrlOperationsResourceUnknown
          Implementations that can distinguish several remote error types
          beyond indication a general ``UrlOperationsRemoteError``:
          ``UrlOperationsInteractionError`` general issues in communicating
          with the remote side; ``UrlOperationsAuthenticationError`` for
          errors related to (failed) authentication at the remote;
          ``UrlOperationsAuthorizationError`` for (lack of) authorizating
          to access a particular resource of perform a particular operation;
          ``UrlOperationsResourceUnknown`` if the target of an operation
          does not exist.
        TimeoutError
          If `timeout` is given and the operation does not complete within the
          number of seconds that a specified by `timeout`.
        """
        raise NotImplementedError

    def upload(self,
               from_path: Path | None,
               to_url: str,
               *,
               credential: str | None = None,
               hash: list[str] | None = None,
               timeout: float | None = None) -> Dict:
        """Upload from a local file or stream to a URL

        Whenever possible, uploads are performed atomically This means that the
        destination will never see a partially uploaded file. It will either
        see the previous content (or nothing) or the newly uploaded content.
        Note: this is not supported by all implementations of URL-operations.

        Parameters
        ----------
        from_path: Path or None
          A local platform-native path or `None`. If `None` the upload
          data is read from `stdin`, otherwise it is read from a file
          at the given path.
        to_url: str
          Valid URL with any scheme supported by a particular implementation.
          The target is assumed to not conflict with existing content, and
          may be overwritten.
        credential: str, optional
          The name of a dedicated credential to be used for authentication
          in order to perform the upload. Particular implementations may
          or may not require or support authentication. They also may or
          may not support automatic credential lookup.
        hash: list(algorithm_names), optional
          If given, must be a list of hash algorithm names supported by the
          `hashlib` module. A corresponding hash will be computed simultaneously
          to the upload (without reading the data twice), and included
          in the return value.
        timeout: float, optional
          If given, specifies a timeout in seconds. If the operation is not
          completed within this time, it will raise a `TimeoutError`-exception.
          If timeout is None, the operation will never timeout.

        Returns
        -------
        dict
          A mapping of property names to values for the completed upload.
          If `hash` algorithm names are provided, a corresponding key for
          each algorithm is included in this mapping, with the hexdigest
          of the corresponding checksum as the value.

        Raises
        ------
        FileNotFoundError
          If the source file cannot be found.
        UrlOperationsRemoteError
          This exception is raised on any deletion-related error on the remote
          side, with a summary of the underlying issues as its message.
          It may carry a status code (e.g. HTTP status code) as its
          ``status_code`` property.
          Any underlying exception must be linked via the `__cause__`
          property (e.g. `raise UrlOperationsRemoteError(...) from ...`).
        UrlOperationsInteractionError
        UrlOperationsAuthenticationError
        UrlOperationsAuthorizationError
        UrlOperationsResourceUnknown
          Implementations that can distinguish several remote error types
          beyond indication a general ``UrlOperationsRemoteError``:
          ``UrlOperationsInteractionError`` general issues in communicating
          with the remote side; ``UrlOperationsAuthenticationError`` for
          errors related to (failed) authentication at the remote;
          ``UrlOperationsAuthorizationError`` for (lack of) authorization
          to access a particular resource of perform a particular operation;
          ``UrlOperationsResourceUnknown`` if the target of an operation
          does not exist.
        TimeoutError
          If `timeout` is given and the operation does not complete within the
          number of seconds that a specified by `timeout`.
        """
        raise NotImplementedError

    def delete(self,
               url: str,
               *,
               credential: str | None = None,
               timeout: float | None = None) -> Dict:
        """Delete a resource identified by a URL

        Parameters
        ----------
        url: str
          Valid URL with any scheme supported by a particular implementation.
        credential: str, optional
          The name of a dedicated credential to be used for authentication
          in order to perform the deletion. Particular implementations may
          or may not require or support authentication. They also may or
          may not support automatic credential lookup.
        timeout: float, optional
          If given, specifies a timeout in seconds. If the operation is not
          completed within this time, it will raise a `TimeoutError`-exception.
          If timeout is None, the operation will never timeout.

        Returns
        -------
        dict
          A mapping of property names to values for the deletion.

        Raises
        ------
        UrlOperationsRemoteError
          This exception is raised on any deletion-related error on the remote
          side, with a summary of the underlying issues as its message.
          It may carry a status code (e.g. HTTP status code) as its
          ``status_code`` property.
          Any underlying exception must be linked via the `__cause__`
          property (e.g. `raise UrlOperationsRemoteError(...) from ...`).
        UrlOperationsInteractionError
        UrlOperationsAuthenticationError
        UrlOperationsAuthorizationError
        UrlOperationsResourceUnknown
          Implementations that can distinguish several remote error types
          beyond indication a general ``UrlOperationsRemoteError``:
          ``UrlOperationsInteractionError`` general issues in communicating
          with the remote side; ``UrlOperationsAuthenticationError`` for
          errors related to (failed) authentication at the remote;
          ``UrlOperationsAuthorizationError`` for (lack of) authorizating
          to access a particular resource of perform a particular operation;
          ``UrlOperationsResourceUnknown`` if the target of an operation
          does not exist.
        TimeoutError
          If `timeout` is given and the operation does not complete within the
          number of seconds that a specified by `timeout`.
        """
        raise NotImplementedError


    def _get_progress_id(self, from_id: str, to_id: str):
        return f'progress_transport_{from_id}_{to_id}'

    def _progress_report_start(self,
                               pid: str,
                               log_msg: tuple,
                               label: str,
                               expected_size: int | None):
        log_progress(
            lgr.info,
            pid,
            *log_msg,
            unit=' Bytes',
            label=label,
            total=expected_size,
            noninteractive_level=logging.DEBUG,
        )

    def _progress_report_update(self, pid: str, log_msg: tuple, increment: int):
        log_progress(
            lgr.info, pid, *log_msg,
            update=increment,
            increment=True,
            noninteractive_level=logging.DEBUG,
        )

    def _progress_report_stop(self, pid: str, log_msg: tuple):
        log_progress(
            lgr.info, pid, *log_msg,
            noninteractive_level=logging.DEBUG,
        )

    def _get_hasher(self, hash: list[str] | None) -> NoOpHash | MultiHash:
        return MultiHash(hash) if hash is not None else NoOpHash()

    def _with_progress(self,
                       stream: Iterable[Any],
                       *,
                       progress_id: str,
                       label: str,
                       expected_size: int | None,
                       start_log_msg: tuple,
                       end_log_msg: tuple,
                       update_log_msg: tuple
                       ) -> Generator[Any, None, None]:
        yield from side_effect(
            lambda chunk: self._progress_report_update(
                progress_id,
                update_log_msg,
                len(chunk)
            ),
            stream,
            before=partial(
                self._progress_report_start,
                progress_id,
                start_log_msg,
                label,
                expected_size
            ),
            after=partial(
                self._progress_report_stop,
                progress_id,
                end_log_msg
            )
        )
